"""fixture used to test DAS model"""
from __future__ import annotations

import numpy as np
import pytest
import torch

from .test_data import NETS_TEST_DATA_ROOT


DAS_TEST_DATA_ROOT = NETS_TEST_DATA_ROOT / "das"

DAS_BATCH_DATA_ROOT = DAS_TEST_DATA_ROOT / "batches"

AMP_TO_DB_IN_BATCHES = [
    batch
    for batch in DAS_BATCH_DATA_ROOT.iterdir()
    if 'amp-to-db-input' in str(batch)
]

AMP_TO_DB_OUT_BATCHES = [
    batch
    for batch in DAS_BATCH_DATA_ROOT.iterdir()
    if 'amp-to-db-output' in str(batch)
]

AMP_TO_DB_IN_OUT_TUPLES = list(zip(AMP_TO_DB_IN_BATCHES, AMP_TO_DB_OUT_BATCHES))


@pytest.fixture(params=AMP_TO_DB_IN_OUT_TUPLES)
def amp_to_db_in_out_tuples(request):
    amp_in_path, amp_out_path = request.param
    amp_in = np.load(amp_in_path)
    amp_out = np.load(amp_out_path)
    return amp_in, amp_out


STFT_KERNELS_ROOT = DAS_TEST_DATA_ROOT / "stft_kernels"
NPY_KERNEL_PATHS = STFT_KERNELS_ROOT.glob('*npy')


@pytest.fixture
def specific_stft_kernels_factory():

    def _specific_stft_kernels_factory(nb_pre_dft):
        assert isinstance(nb_pre_dft, int), f"nb_pre_dft should be an integer but was: {nb_pre_dft}"

        these = [
            path
            for path in NPY_KERNEL_PATHS
            if str(nb_pre_dft) in str(path)
        ]

        err = f"did not find two npy kernel paths for nb_pre_dft={nb_pre_dft}, found: {these}"
        assert len(these) == 2, err

        real_path = [path for path in these if 'real' in str(path)]
        imag_path = [path for path in these if 'imag' in str(path)]

        assert len(real_path) == 1, f'did not find single path for real kernel, found: {real}'
        real_path = real_path[0]

        assert len(imag_path) == 1, f'did not find single path for img kernel, found: {img}'
        imag_path = imag_path[0]

        real = np.load(real_path)
        imag = np.load(imag_path)

        return real, imag

    return _specific_stft_kernels_factory


INPUTS_BATCHES = [
    batch
    for batch in DAS_BATCH_DATA_ROOT.iterdir()
    if '-inputs-' in str(batch)
]

CONV2D_OUTPUT_REAL_BATCHES = [
    batch
    for batch in DAS_BATCH_DATA_ROOT.iterdir()
    if 'conv2d-output-real-' in str(batch)
]

CONV2D_OUTPUT_IMAG_BATCHES = [
    batch
    for batch in DAS_BATCH_DATA_ROOT.iterdir()
    if 'conv2d-output-imag-' in str(batch)
]

INPUTS_CONV2D_OUTPUTS_TUPLES = list(zip(
    INPUTS_BATCHES, CONV2D_OUTPUT_REAL_BATCHES, CONV2D_OUTPUT_IMAG_BATCHES)
)


@pytest.fixture(params=INPUTS_CONV2D_OUTPUTS_TUPLES)
def inputs_conv2d_outputs(request):
    inputs_path, conv2d_real_path, conv2d_imag_path = request.param
    inputs = torch.tensor(
        np.load(inputs_path)
    ).to(torch.float32)  # audio is 'half'
    conv2d_real = torch.tensor(
        np.load(conv2d_real_path)
    )
    conv2d_imag = torch.tensor(
        np.load(conv2d_imag_path)
    )
    return inputs, conv2d_real, conv2d_imag


STFT_OUT_BATCHES = [
    batch
    for batch in DAS_BATCH_DATA_ROOT.iterdir()
    if str(batch.name).startswith('stft-layer-out-')
]

INPUTS_STFT_OUT_TUPLES = list(zip(INPUTS_BATCHES, STFT_OUT_BATCHES))


@pytest.fixture(params=INPUTS_STFT_OUT_TUPLES)
def inputs_stft_out_tuples(request):
    inputs_path, stft_path = request.param
    inputs = torch.tensor(
        np.load(inputs_path)
    ).to(torch.float32)  # audio is 'half'
    stft_out = torch.tensor(
        np.load(stft_path)
    )
    return inputs, stft_out



"""
"""

from __future__ import annotations

import torch

from .. import metrics, nets
from .decorator import model
from .frame_classification_model import FrameClassificationModel


@model(family=FrameClassificationModel)
class ED_TCN:
    """Encoder-Decoder Temporal Convolutional Network model
    for the frame classification task.
    As described in [1]_. This adaptation adds convolutional
    blocks as a front end to extract features from the input.

    Attributes
    ----------
    network : vak.nets.ED_TCN
        Encoder-Decoder Temporal Convolutional Network architecture.
    loss: torch.nn.CrossEntropyLoss
        Standard cross-entropy loss
    optimizer: torch.optim.Adam
        Adam optimizer.
    metrics: dict
        Mapping string names to the following metrics:
        ``vak.metrics.Accuracy``, ``vak.metrics.Levenshtein``,
        ``vak.metrics.CharacterErrorRate``, ``torch.nn.CrossEntropyLoss``.

    References
    ----------
    .. [1] Lea, C., Flynn, M. D., Vidal, R., Reiter, A., & Hager, G. D. (2017).
       Temporal convolutional networks for action segmentation and detection.
       In proceedings of the IEEE Conference on Computer Vision and Pattern Recognition (pp. 156-165).
    """

    network = nets.ED_TCN
    loss = torch.nn.CrossEntropyLoss
    optimizer = torch.optim.Adam
    metrics = {
        "acc": metrics.Accuracy,
        "levenshtein": metrics.Levenshtein,
        "character_error_rate": metrics.CharacterErrorRate,
        "loss": torch.nn.CrossEntropyLoss,
    }
    default_config = {"optimizer": {"lr": 0.003}}



from . import (
    audio_dataset,
    constants,
    dataset_df_helper,
    frame_classification,
    parametric_umap,
    prep_,
    sequence_dataset,
    spectrogram_dataset,
    unit_dataset,
)
from .prep_ import prep

__all__ = [
    "audio_dataset",
    "constants",
    "dataset_df_helper",
    "frame_classification",
    "parametric_umap",
    "prep",
    "prep_",
    "sequence_dataset",
    "spectrogram_dataset",
    "unit_dataset",
]



"""Function that generates new inferences from trained models in the frame classification family."""

from __future__ import annotations

import logging
import os
import pathlib

import lightning
import torch.utils.data

from .. import datapipes, models
from ..common import validators
from ..datapipes.parametric_umap import Datapipe

logger = logging.getLogger(__name__)


def predict_with_parametric_umap_model(
    model_config: dict,
    dataset_config: dict,
    trainer_config: dict,
    checkpoint_path,
    num_workers=2,
    transform_params: dict | None = None,
    output_dir=None,
):
    """Make predictions on a dataset with a trained
    :class:`vak.models.ParametricUMAPModel`.

     Parameters
     ----------
    model_config : dict
        Model configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.ModelConfig.asdict`.
    dataset_config: dict
        Dataset configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.DatasetConfig.asdict`.
    trainer_config: dict
        Configuration for :class:`lightning.pytorch.Trainer`.
        Can be obtained by calling :meth:`vak.config.TrainerConfig.asdict`.
    checkpoint_path : str
        path to directory with checkpoint files saved by Torch, to reload model
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader. Default is 2.
    timebins_key : str
        key for accessing vector of time bins in files. Default is 't'.
    annot_csv_filename : str
        name of .csv file containing predicted annotations.
        Default is None, in which case the name of the dataset .csv
        is used, with '.annot.csv' appended to it.
    output_dir : str, Path
        path to location where .csv containing predicted annotation
        should be saved. Defaults to current working directory.
    """
    for path, path_name in zip(
        (checkpoint_path,),
        ("checkpoint_path",),
    ):
        if path is not None:
            if not validators.is_a_file(path):
                raise FileNotFoundError(
                    f"value for ``{path_name}`` not recognized as a file: {path}"
                )

    dataset_path = pathlib.Path(dataset_config["path"])
    if not dataset_path.exists() or not dataset_path.is_dir():
        raise NotADirectoryError(
            f"`dataset_path` not found or not recognized as a directory: {dataset_path}"
        )
    logger.info(
        f"Loading metadata from dataset path: {dataset_path}",
    )
    metadata = datapipes.frame_classification.Metadata.from_dataset_path(
        dataset_path
    )

    if output_dir is None:
        output_dir = pathlib.Path(os.getcwd())
    else:
        output_dir = pathlib.Path(output_dir)

    if not output_dir.is_dir():
        raise NotADirectoryError(
            f"value specified for output_dir is not recognized as a directory: {output_dir}"
        )

    # ---------------- load data for prediction ------------------------------------------------------------------------
    model_name = model_config["name"]

    dataset_csv_path = dataset_path / metadata.dataset_csv_filename
    logger.info(
        f"loading dataset to predict from csv path: {dataset_csv_path}"
    )

    pred_dataset = Datapipe.from_dataset_path(
        dataset_path=dataset_path,
        split="predict",
        **dataset_config["params"],
    )

    pred_loader = torch.utils.data.DataLoader(
        dataset=pred_dataset,
        shuffle=False,
        # batch size 1 because each spectrogram reshaped into a batch of windows
        batch_size=1,
        num_workers=num_workers,
    )

    # ---------------- do the actual predicting + converting to annotations --------------------------------------------
    input_shape = pred_dataset.shape
    # if dataset returns spectrogram reshaped into windows,
    # throw out the window dimension; just want to tell network (channels, height, width) shape
    if len(input_shape) == 4:
        input_shape = input_shape[1:]
    logger.info(
        f"Shape of input to networks used for predictions: {input_shape}"
    )

    logger.info(f"instantiating model from config:/n{model_name}")

    model = models.get(
        model_name,
        model_config,
        input_shape=input_shape,
    )

    # ---------------- do the actual predicting --------------------------------------------------------------------
    logger.info(
        f"loading checkpoint for {model_name} from path: {checkpoint_path}"
    )
    model.load_state_dict_from_path(checkpoint_path)

    trainer_logger = lightning.pytorch.loggers.TensorBoardLogger(
        save_dir=output_dir
    )
    trainer = lightning.pytorch.Trainer(
        accelerator=trainer_config["accelerator"],
        devices=trainer_config["devices"],
        logger=trainer_logger,
    )

    logger.info(f"running predict method of {model_name}")
    results = trainer.predict(model, pred_loader)  # noqa : F841

    # eval_df = pd.DataFrame(row, index=[0])
    # eval_csv_path = output_dir.joinpath(f"eval_{model_name}_{timenow}.csv")
    # logger.info(f"saving csv with evaluation metrics at: {eval_csv_path}")
    # eval_df.to_csv(
    #     eval_csv_path, index=False
    # )  # index is False to avoid having "Unnamed: 0" column when loading



"""Registry for models.

Makes it possible to register a model declared outside of ``vak``
with a decorator, so that the model can be used at runtime.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Type

import lightning

if TYPE_CHECKING:
    from .factory import ModelFactory

MODEL_FAMILY_REGISTRY = {}


def model_family(family_class: Type) -> None:
    """Decorator that adds a :class:`lightning.LightningModule` class to the registry of model families."""
    if not issubclass(family_class, lightning.LightningModule):
        raise TypeError(
            "The ``family_class`` provided to the `vak.models.model_family` decorator"
            "must be a subclass of `lightning.LightningModule`, "
            f"but the class specified is not: {family_class}. "
            f"Subclasses of `lightning.LightningModule` are: {lightning.LightningModule.__subclasses__()}"
        )

    model_family_name = family_class.__name__
    if model_family_name in MODEL_FAMILY_REGISTRY:
        raise ValueError(
            f"Attempted to register a model family with the name '{model_family_name}', "
            f"but this name is already in the registry:\n{MODEL_FAMILY_REGISTRY}"
        )

    MODEL_FAMILY_REGISTRY[model_family_name] = family_class
    # need to return class after we register it or we replace it with None
    # when this function is used as a decorator
    return family_class


MODEL_REGISTRY = {}


def register_model(model: ModelFactory) -> ModelFactory:
    """Function that registers a model in the model registry.

    This function is called by :func:`vak.models.decorator.model`,
    that creates an instance of a :class:`vak.models.ModelFactory`,
    given a :class:`vak.models.definition.ModelDefinition`
    and a :class:`lightning.LightningModule` class that has been
    registered as a model family with :func:`model_family`.

    So you will not usually need to use this function directly,
    and should prefer to use :func:`vak.models.decorator.model` instead.
    """
    model_family_classes = list(MODEL_FAMILY_REGISTRY.values())
    model_family = model.family
    if model_family not in model_family_classes:
        raise TypeError(
            "The family of `model` passed to the `register_model` decorator "
            f"is not recognized as a model family. Class was '{model}' and "
            f"its family is '{model_family}'. "
            f"Please specify a valid model family. "
            f"Valid model family classes are: {model_family_classes}"
        )

    model_name = model.__name__
    if model_name in MODEL_REGISTRY:
        raise ValueError(
            f"Attempted to register a model family with the name '{model_name}', "
            f"but this name is already in the registry.\n"
        )

    MODEL_REGISTRY[model_name] = model
    # need to return class after we register it,
    # or we would replace it with None when this function is used as a decorator
    return model


def __getattr__(name: str) -> Any:
    """Module-level __getattr__ function that we use to dynamically determine models."""
    if name == "MODEL_FAMILY_FROM_NAME":
        return {
            model_name: model.family.__name__
            for model_name, model in MODEL_REGISTRY.items()
        }
    elif name == "MODEL_NAMES":
        return list(MODEL_REGISTRY.keys())
    else:
        raise AttributeError(
            f"Not an attribute of `vak.models.registry`: {name}"
        )



import logging
import shutil
from pathlib import Path

from .. import config
from .. import train as train_module
from ..common.logging import config_logging_for_cli, log_version
from ..common.paths import generate_results_dir_name_as_path

logger = logging.getLogger(__name__)


def train(toml_path):
    """train models using training set specified in config.toml file.
    Function called by command-line interface.

    Trains models, saves results in new directory within root_results_dir specified
    in config.toml file, and adds path to that new directory to config.toml file.

    Parameters
    ----------
    toml_path : str, Path
        path to a configuration file in TOML format.
    """
    toml_path = Path(toml_path)
    cfg = config.Config.from_toml_path(toml_path)

    if cfg.train is None:
        raise ValueError(
            f"train called with a config.toml file that does not have a TRAIN section: {toml_path}"
        )

    # ---- set up directory to save output -----------------------------------------------------------------------------
    results_path = generate_results_dir_name_as_path(
        cfg.train.root_results_dir
    )
    results_path.mkdir(parents=True)
    # copy config file into results dir now that we've made the dir
    shutil.copy(toml_path, results_path)

    # ---- set up logging ----------------------------------------------------------------------------------------------
    config_logging_for_cli(
        log_dst=results_path, log_stem="train", level="INFO", force=True
    )
    log_version(logger)
    logger.info("Logging results to {}".format(results_path))

    if cfg.train.dataset.path is None:
        raise ValueError(
            "No value is specified for 'dataset_path' in this .toml config file."
            f"To generate a .csv file that represents the dataset, "
            f"please run the following command:\n'vak prep {toml_path}'"
        )

    train_module.train(
        model_config=cfg.train.model.asdict(),
        dataset_config=cfg.train.dataset.asdict(),
        trainer_config=cfg.train.trainer.asdict(),
        batch_size=cfg.train.batch_size,
        num_epochs=cfg.train.num_epochs,
        num_workers=cfg.train.num_workers,
        checkpoint_path=cfg.train.checkpoint_path,
        frames_standardizer_path=cfg.train.frames_standardizer_path,
        results_path=results_path,
        standardize_frames=cfg.train.standardize_frames,
        shuffle=cfg.train.shuffle,
        val_step=cfg.train.val_step,
        ckpt_step=cfg.train.ckpt_step,
        patience=cfg.train.patience,
    )



import torch


class CrossEntropyLoss(torch.nn.CrossEntropyLoss):
    """Wrapper around :class:`torch.nn.CrossEntropyLoss`

    Converts the argument ``weight`` to a :class:`torch.Tensor`
    if it is a :class:`list`.
    """

    def __init__(
        self,
        weight=None,
        size_average=None,
        ignore_index=-100,
        reduce=None,
        reduction="mean",
        label_smoothing=0.0,
    ):

        if weight is not None:
            if isinstance(weight, torch.Tensor):
                pass
            elif isinstance(weight, list):
                weight = torch.Tensor(weight)
        super().__init__(
            weight,
            size_average,
            ignore_index,
            reduce,
            reduction,
            label_smoothing,
        )



import json
import math

import numpy as np
import pandas as pd
import pytest

import vak.common  # for constants
import vak.common.annotation
import vak.prep.spectrogram_dataset.spect_helper
import vak.prep.split.split


NUM_SAMPLES = 10  # number of times to sample behavior of random-number generator


def train_test_dur_split_inds_output_matches_expected(
    train_dur,
    val_dur,
    test_dur,
    labelset,
    durs,
    labels,
    train_inds,
    val_inds,
    test_inds,
):
    for split, dur_in, inds in zip(
        ("train", "val", "test"),
        (train_dur, val_dur, test_dur),
        (train_inds, val_inds, test_inds),
    ):
        if dur_in is not None:
            dur_out = sum([durs[ind] for ind in inds])
            if dur_in >= 0:
                assert dur_out >= dur_in
            elif dur_in == -1:
                if split == "train":
                    assert math.isclose(
                        dur_out, sum(durs) - sum([durs[ind] for ind in test_inds])
                    )
                elif split == "test":
                    assert math.isclose(
                        dur_out, sum(durs) - sum([durs[ind] for ind in train_inds])
                    )

            all_lbls_this_set = [lbl for ind in inds for lbl in labels[ind]]
            assert labelset == set(all_lbls_this_set)
        else:
            assert inds is None

    assert set(train_inds).isdisjoint(set(test_inds))
    if val_dur is not None:
        assert set(train_inds).isdisjoint(set(val_inds))
        assert set(test_inds).isdisjoint(set(val_inds))

    return True


@pytest.mark.parametrize(
    "durs, labels, labelset, train_dur, val_dur, test_dur",
    [
        (
            (5, 5, 5, 5, 5),
            ([np.asarray(list("abcde")) for _ in range(5)]),
            set(list("abcde")),
            20,
            None,
            5,
        ),
        (
            (3, 2, 1, 3, 2, 3, 2, 1, 3, 2),
            ["abc", "ab", "c", "cde", "de", "abc", "ab", "c", "cde", "de"],
            set(list("abcde")),
            14,
            None,
            8,
        ),
        (
            (3, 2, 1, 3, 2, 3, 2, 1, 3, 2),
            ["abc", "ab", "c", "cde", "de", "abc", "ab", "c", "cde", "de"],
            set(list("abcde")),
            8,
            7,
            7,
        ),
    ],
)
def test_train_test_dur_split_inds_fake_data(
    durs, labels, labelset, train_dur, val_dur, test_dur
):
    for _ in range(NUM_SAMPLES):
        train_inds, val_inds, test_inds = vak.prep.split.split.train_test_dur_split_inds(
            durs, labels, labelset, train_dur, test_dur, val_dur
        )

        assert train_test_dur_split_inds_output_matches_expected(
            train_dur,
            val_dur,
            test_dur,
            labelset,
            durs,
            labels,
            train_inds,
            val_inds,
            test_inds,
        )


def test_train_test_dur_split_inds_fake_data_impossible():
    durs = (3, 2, 1, 3, 2, 3, 2, 1, 3, 2)
    labelset = set(list("abcde"))
    labels = ["abc", "ab", "c", "cde", "de", "abc", "ab", "c", "cde", "de"]
    labels = [np.asarray(list(lbl)) for lbl in labels]
    train_dur = 16
    val_dur = 2
    test_dur = 4
    with pytest.raises(ValueError):
        vak.prep.split.split.train_test_dur_split_inds(
            durs, labels, labelset, train_dur, test_dur, val_dur
        )


@pytest.mark.parametrize(
    "train_dur, val_dur, test_dur",
    [
        (35, 20, 35),
        (35, None, -1),
        (-1, None, 35),
    ],
)
def test_train_test_dur_split_inds_audio_cbin_annot_notmat(
    train_dur, val_dur, test_dur, audio_cbin_annot_notmat_durs_labels, labelset_notmat
):
    labelset_notmat = set(labelset_notmat)
    durs, labels = audio_cbin_annot_notmat_durs_labels

    for _ in range(NUM_SAMPLES):
        train_inds, val_inds, test_inds = vak.prep.split.split.train_test_dur_split_inds(
            durs, labels, labelset_notmat, train_dur, test_dur, val_dur
        )

        assert train_test_dur_split_inds_output_matches_expected(
            train_dur,
            val_dur,
            test_dur,
            labelset_notmat,
            durs,
            labels,
            train_inds,
            val_inds,
            test_inds,
        )


@pytest.mark.parametrize(
    "train_dur, val_dur, test_dur",
    [(200, 100, 200), (200, None, -1), (-1, None, 200), (200, 100, 200)],
)
def test_train_test_dur_split_inds_spect_mat(
    train_dur, val_dur, test_dur, spect_mat_annot_yarden_durs_labels, labelset_yarden
):
    labelset_yarden = set(labelset_yarden)
    durs, labels = spect_mat_annot_yarden_durs_labels
    train_dur = 200
    val_dur = 100
    test_dur = 200

    for _ in range(NUM_SAMPLES):
        train_inds, val_inds, test_inds = vak.prep.split.split.train_test_dur_split_inds(
            durs, labels, labelset_yarden, train_dur, test_dur, val_dur
        )

        assert train_test_dur_split_inds_output_matches_expected(
            train_dur,
            val_dur,
            test_dur,
            labelset_yarden,
            durs,
            labels,
            train_inds,
            val_inds,
            test_inds,
        )


def test_dataframe_None_raises():
    durs = (5, 5, 5, 5, 5)
    labelset = set(list("abcde"))
    labels = [np.asarray(list(labelset)) for _ in range(5)]

    train_dur = None
    val_dur = None
    test_dur = None

    with pytest.raises(ValueError):
        vak.prep.split.split.train_test_dur_split_inds(
            durs, labels, labelset, train_dur, test_dur, val_dur
        )


def test_dataframe_only_val_raises():
    durs = (5, 5, 5, 5, 5)
    labelset = set(list("abcde"))
    labels = [np.asarray(list(labelset)) for _ in range(5)]

    train_dur = None
    val_dur = 100
    test_dur = None

    # because we only specified duration for validation set
    with pytest.raises(ValueError):
        vak.prep.split.split.train_test_dur_split_inds(
            durs, labels, labelset, train_dur, test_dur, val_dur
        )


def test_dataframe_negative_dur_raises():
    durs = (5, 5, 5, 5, 5)
    labelset = set(list("abcde"))
    labels = [np.asarray(list(labelset)) for _ in range(5)]

    train_dur = -2
    test_dur = None
    val_dur = 100

    # because negative duration is invalid
    with pytest.raises(ValueError):
        vak.prep.split.split.train_test_dur_split_inds(
            durs, labels, labelset, train_dur, test_dur, val_dur
        )


def test_dataframe_specd_dur_gt_raises():
    durs = (5, 5, 5, 5, 5)
    labelset = set(list("abcde"))
    labels = [np.asarray(list(labelset)) for _ in range(5)]

    train_dur = 100
    test_dur = 100
    val_dur = 100
    # because total splits duration is greater than dataset duration
    with pytest.raises(ValueError):
        vak.prep.split.split.train_test_dur_split_inds(
            durs, labels, labelset, train_dur, test_dur, val_dur
        )


@pytest.mark.parametrize(
    "config_type, model_name, spect_format, audio_format, annot_format, train_dur, val_dur, test_dur",
    [
        ('train', 'TweetyNet', None, 'cbin', 'notmat', 45, None, None,),
        ('train', 'TweetyNet', None, 'cbin', 'notmat', 45, None, 30,),
        ('train', 'TweetyNet', None, 'cbin', 'notmat', 45, 15, 30,),
        ('train', 'TweetyNet', None, 'cbin', 'notmat', None, None, 30,),
        ('train', 'TweetyNet', 'mat', None, 'yarden', 200, None, None,),
        ('train', 'TweetyNet', 'mat', None, 'yarden', None, None, 200,),
        ('train', 'TweetyNet', 'mat', None, 'yarden', 200, None, 200,),
        ('train', 'TweetyNet', 'mat', None, 'yarden', 200, 80, 120,),
    ]
)
def test_split_frame_classification_dataframe(
    config_type, model_name, spect_format, audio_format, annot_format,  specific_dataset_path,
    train_dur, val_dur, test_dur,
):
    dataset_path = specific_dataset_path(
        spect_format=spect_format,
        audio_format=audio_format,
        annot_format=annot_format,
        config_type=config_type,
        model=model_name,
    )
    metadata = vak.datapipes.frame_classification.Metadata.from_dataset_path(dataset_path)
    dataset_csv_path = dataset_path / metadata.dataset_csv_filename
    dataset_df = pd.read_csv(dataset_csv_path)
    dataset_df = dataset_df.drop(columns=('split'))
    labelmap_path = dataset_path / "labelmap.json"
    with labelmap_path.open("r") as f:
        labelmap = json.load(f)
    labelset = set(key for key in labelmap.keys() if key != vak.common.constants.DEFAULT_BACKGROUND_LABEL)

    dataset_df_split = vak.prep.split.split.frame_classification_dataframe(
        dataset_df, dataset_path, labelset=labelset, train_dur=train_dur, val_dur=val_dur, test_dur=test_dur
    )

    assert isinstance(dataset_df_split, pd.DataFrame)

    for split, duration in zip(
            ('train', 'val', 'test'),
            (train_dur, val_dur, test_dur),
    ):
        if duration is not None:
            duration_out = dataset_df_split[dataset_df_split["split"] == split].duration.sum()
            assert duration_out >= duration
        else:
            assert split not in dataset_df_split["split"].unique().tolist()


@pytest.mark.parametrize(
    "config_type, model_name, spect_format, audio_format, annot_format, train_dur, val_dur, test_dur",
    [
        ('train', 'ConvEncoderUMAP', None, 'cbin', 'notmat', 0.2, None, None,),
        ('train', 'ConvEncoderUMAP', None, 'cbin', 'notmat', 0.2, None, 0.15,),
        ('train', 'ConvEncoderUMAP', None, 'cbin', 'notmat', 0.2, 0.1, 0.15,),
    ]
)
def test_split_unit_dataframe(
    config_type, model_name, spect_format, audio_format, annot_format,  specific_dataset_path,
    train_dur, val_dur, test_dur,
):
    dataset_path = specific_dataset_path(
        spect_format=spect_format,
        audio_format=audio_format,
        annot_format=annot_format,
        config_type=config_type,
        model=model_name,
    )
    metadata = vak.datapipes.parametric_umap.Metadata.from_dataset_path(dataset_path)
    dataset_csv_path = dataset_path / metadata.dataset_csv_filename
    dataset_df = pd.read_csv(dataset_csv_path)
    dataset_df = dataset_df.drop(columns=('split'))
    labelmap_path = dataset_path / "labelmap.json"
    with labelmap_path.open("r") as f:
        labelmap = json.load(f)
    labelset = set(key for key in labelmap.keys() if key != vak.common.constants.DEFAULT_BACKGROUND_LABEL)

    dataset_df_split = vak.prep.split.split.unit_dataframe(
        dataset_df, dataset_path, labelset=labelset, train_dur=train_dur, val_dur=val_dur, test_dur=test_dur
    )

    assert isinstance(dataset_df_split, pd.DataFrame)

    for split, duration in zip(
            ('train', 'val', 'test'),
            (train_dur, val_dur, test_dur),
    ):
        if duration is not None:
            duration_out = dataset_df_split[dataset_df_split["split"] == split].duration.sum()
            assert duration_out >= duration
        else:
            assert split not in dataset_df_split["split"].unique().tolist()


"""Class and functions for ``[vak.prep]`` table of configuration file."""

from __future__ import annotations

import inspect

import dask.bag
from attrs import converters, define, field, validators
from attrs.validators import instance_of

from .. import prep
from ..common.converters import expanded_user_path, labelset_to_set
from .spect_params import SpectParamsConfig
from .validators import is_annot_format, is_audio_format, is_spect_format


def duration_from_toml_value(value):
    """converter for dataset split durations.
    If value is -1, that value is returned -- specifies "use the remainder of the dataset".
    Other values are converted to float when possible."""
    if value == -1:
        return value
    else:
        return float(value)


def is_valid_duration(instance, attribute, value):
    """validator for dataset split durations"""
    if type(value) not in {int, float}:
        raise TypeError(
            f"invalid type for {attribute} of {instance}: {type(value)}. Type should be float or int."
        )

    if value == -1:  # specifies "use the remainder of the dataset"
        # so it is valid, but other negative values are not
        return

    if not value >= 0:
        raise ValueError(
            f"value specified for {attribute} of {instance} must be greater than or equal to zero, was {value}"
        )


def are_valid_dask_bag_kwargs(instance, attribute, value):
    """validator for ``audio_dask_bag_kwargs``"""
    if not isinstance(value, dict):
        raise TypeError(
            f"Option ``audio_dask_bag_kwargs`` should be a dict but was a {type(value)}.\n"
            "So that it parses as a dict, please specify this option "
            "as an inline table in the .toml file, e.g.\n"
            "`audio_dask_bag_kwargs = { npartitions = 20}`"
        )
    kwargs = list(value.keys())
    valid_bag_kwargs = list(
        inspect.signature(dask.bag.from_sequence).parameters.keys()
    )
    if not all([kwarg in valid_bag_kwargs for kwarg in kwargs]):
        invalid_kwargs = [
            kwarg for kwarg in kwargs if kwarg not in valid_bag_kwargs
        ]
        print(
            f"Invalid keyword arguments specified in ``audio_dask_bag_kwargs``: {invalid_kwargs}"
        )


REQUIRED_KEYS = (
    "data_dir",
    "output_dir",
)


@define
class PrepConfig:
    """Class that represents ``[vak.prep]`` table of configuration file.

    Attributes
    ----------
    data_dir : str
        path to directory with files from which to make dataset
    output_dir : str
        Path to location where data sets should be saved. Default is None,
        in which case data sets are saved in the current working directory.
    dataset_type : str
        String name of the type of dataset, e.g.,
        'frame_classification'. Dataset types are
        defined by machine learning tasks, e.g.,
        a 'frame_classification' dataset would be used
        a :class:`vak.models.FrameClassificationModel` model.
        Valid dataset types are defined as
        :const:`vak.prep.prep.DATASET_TYPES`.
    audio_format : str
        format of audio files. One of {'wav', 'cbin'}.
    spect_format : str
        format of files containg spectrograms as 2-d matrices.
        One of {'mat', 'npy'}.
    spect_params: vak.config.SpectParamsConfig, optional
        Parameters for Short-Time Fourier Transform and post-processing
        of spectrograms.
        Instance of :class:`vak.config.SpectParamsConfig` class.
        Optional, default is None.
    annot_format : str
        format of annotations. Any format that can be used with the
        crowsetta library is valid.
    annot_file : str
        Path to a single annotation file. Default is None.
        Used when a single file contains annotations for multiple audio files.
    labelset : set
        of str or int, the set of labels that correspond to annotated segments
        that a network should learn to segment and classify. Note that if there
        are segments that are not annotated, e.g. silent gaps between songbird
        syllables, then `vak` will assign a dummy label to those segments
        -- you don't have to give them a label here.
        Value for ``labelset`` is converted to a Python ``set``
        using ``vak.config.converters.labelset_from_toml_value``.
        See help for that function for details on how to specify labelset.
    audio_dask_bag_kwargs : dict
        Keyword arguments used when calling ``dask.bag.from_sequence``
        inside ``vak.io.audio``, where it is used to parallelize
        the conversion of audio files into spectrograms.
        Option should be specified in config.toml file as an inline table,
        e.g., ``audio_dask_bag_kwargs = { npartitions = 20 }``.
        Allows for finer-grained control
        when needed to process files of different sizes.
    train_dur : float
        total duration of training set, in seconds. When creating a learning curve,
        training subsets of shorter duration (specified by the 'train_set_durs' option
        in the LEARNCURVE section of a config.toml file) will be drawn from this set.
    val_dur : float
        total duration of validation set, in seconds.
    test_dur : float
        total duration of test set, in seconds.
    train_set_durs : list, optional
        Durations of datasets to use for a learning curve.
        Float values, durations in seconds of subsets taken from training data
        to create a learning curve, e.g. [5., 10., 15., 20.]. Default is None.
        Required if config file has a learncurve section.
    num_replicates : int, optional
        Number of replicates to train for each training set duration
        in a learning curve. Each replicate uses a different
        randomly drawn subset of the training data (but of the same duration).
        Default is None. Required if config file has a learncurve section.
    """

    data_dir = field(converter=expanded_user_path)
    output_dir = field(converter=expanded_user_path)

    dataset_type = field(validator=instance_of(str))

    @dataset_type.validator
    def is_valid_dataset_type(self, attribute, value):
        if value not in prep.constants.DATASET_TYPES:
            raise ValueError(
                f"Invalid dataset type: {value}. "
                f"Valid dataset types are: {prep.constants.DATASET_TYPES}"
            )

    input_type = field(validator=instance_of(str))

    @input_type.validator
    def is_valid_input_type(self, attribute, value):
        if value not in prep.constants.INPUT_TYPES:
            raise ValueError(
                f"Invalid input type: {value}. Must be one of: {prep.constants.INPUT_TYPES}"
            )

    audio_format = field(
        validator=validators.optional(is_audio_format), default=None
    )
    spect_format = field(
        validator=validators.optional(is_spect_format), default=None
    )
    spect_params = field(
        validator=validators.optional(instance_of(SpectParamsConfig)),
        default=None,
    )
    annot_file = field(
        converter=converters.optional(expanded_user_path),
        default=None,
    )
    annot_format = field(
        validator=validators.optional(is_annot_format), default=None
    )

    labelset = field(
        converter=converters.optional(labelset_to_set),
        validator=validators.optional(instance_of(set)),
        default=None,
    )

    audio_dask_bag_kwargs = field(
        validator=validators.optional(are_valid_dask_bag_kwargs), default=None
    )

    train_dur = field(
        converter=converters.optional(duration_from_toml_value),
        validator=validators.optional(is_valid_duration),
        default=None,
    )
    val_dur = field(
        converter=converters.optional(duration_from_toml_value),
        validator=validators.optional(is_valid_duration),
        default=None,
    )
    test_dur = field(
        converter=converters.optional(duration_from_toml_value),
        validator=validators.optional(is_valid_duration),
        default=None,
    )
    train_set_durs = field(
        validator=validators.optional(instance_of(list)), default=None
    )
    num_replicates = field(
        validator=validators.optional(instance_of(int)), default=None
    )

    def __attrs_post_init__(self):
        if self.audio_format is not None and self.spect_format is not None:
            raise ValueError("cannot specify audio_format and spect_format")

        if self.audio_format is None and self.spect_format is None:
            raise ValueError(
                "must specify either audio_format or spect_format"
            )

    @classmethod
    def from_config_dict(cls, config_dict: dict) -> PrepConfig:
        """Return :class:`PrepConfig` instance from a :class:`dict`.

        The :class:`dict` passed in should be the one found
        by loading a valid configuration toml file with
        :func:`vak.config.parse.from_toml_path`,
        and then using key ``prep``,
        i.e., ``PrepConfig.from_config_dict(config_dict['prep'])``."""
        for required_key in REQUIRED_KEYS:
            if required_key not in config_dict:
                raise KeyError(
                    "The `[vak.prep]` table in a configuration file requires "
                    f"the key '{required_key}', but it was not found "
                    "when loading the configuration file into a Python dictionary. "
                    "Please check that the configuration file is formatted correctly."
                )
        if "spect_params" in config_dict:
            config_dict["spect_params"] = SpectParamsConfig(
                **config_dict["spect_params"]
            )
        return cls(**config_dict)



"""tests for vak.config.prep module"""
import copy

import pytest

import vak.config.prep


class TestPrepConfig:
    @pytest.mark.parametrize(
        'config_dict',
        [
            {
                'annot_format': 'notmat',
                'audio_format': 'cbin',
                'data_dir': './tests/data_for_tests/source/audio_cbin_annot_notmat/gy6or6/032412',
                'dataset_type': 'parametric umap',
                'input_type': 'spect',
                'labelset': 'iabcdefghjk',
                'output_dir': './tests/data_for_tests/generated/prep/eval/audio_cbin_annot_notmat/ConvEncoderUMAP',
                'spect_params': {'fft_size': 512,
                                'step_size': 32,
                                'transform_type': 'log_spect_plus_one'},
                'test_dur': 0.2
            },
            {
                'annot_format': 'notmat',
                'audio_format': 'cbin',
                'data_dir': './tests/data_for_tests/source/audio_cbin_annot_notmat/gy6or6/032312',
                'dataset_type': 'frame classification',
                'input_type': 'spect',
                'labelset': 'iabcdefghjk',
                'output_dir': './tests/data_for_tests/generated/prep/train/audio_cbin_annot_notmat/TweetyNet',
                'spect_params': {'fft_size': 512,
                                'freq_cutoffs': [500, 10000],
                                'step_size': 64,
                                'thresh': 6.25,
                                'transform_type': 'log_spect'},
                'test_dur': 30,
                'train_dur': 50,
                'val_dur': 15
            },
        ]
    )
    def test_init(self, config_dict):
        config_dict['spect_params'] = vak.config.SpectParamsConfig(**config_dict['spect_params'])

        prep_config = vak.config.PrepConfig(**config_dict)

        assert isinstance(prep_config, vak.config.prep.PrepConfig)
        for key, val in config_dict.items():
            assert hasattr(prep_config, key)
            if key == 'data_dir' or key == 'output_dir':
                assert getattr(prep_config, key) == vak.common.converters.expanded_user_path(val)
            elif key == 'labelset':
                assert getattr(prep_config, key) == vak.common.converters.labelset_to_set(val)
            else:
                assert getattr(prep_config, key) == val

    @pytest.mark.parametrize(
        'config_dict',
            [
                {
                    'annot_format': 'notmat',
                    'audio_format': 'cbin',
                    'spect_format': 'mat',
                    'data_dir': './tests/data_for_tests/source/audio_cbin_annot_notmat/gy6or6/032412',
                    'dataset_type': 'parametric umap',
                    'input_type': 'spect',
                    'labelset': 'iabcdefghjk',
                    'output_dir': './tests/data_for_tests/generated/prep/eval/audio_cbin_annot_notmat/ConvEncoderUMAP',
                    'spect_params': {'fft_size': 512,
                                    'step_size': 32,
                                    'transform_type': 'log_spect_plus_one'},
                    'test_dur': 0.2
                },
        ]
    )
    def test_both_audio_and_spect_format_raises(
        self, config_dict,
    ):
        """test that a config with both an audio and a spect format raises a ValueError"""
        # need to do this set-up so we don't mask one error with another
        config_dict['spect_params'] = vak.config.SpectParamsConfig(**config_dict['spect_params'])

        with pytest.raises(ValueError):
            prep_config = vak.config.PrepConfig(**config_dict)

    @pytest.mark.parametrize(
        'config_dict',
            [
                {
                    'annot_format': 'notmat',
                    'data_dir': './tests/data_for_tests/source/audio_cbin_annot_notmat/gy6or6/032412',
                    'dataset_type': 'parametric umap',
                    'input_type': 'spect',
                    'labelset': 'iabcdefghjk',
                    'output_dir': './tests/data_for_tests/generated/prep/eval/audio_cbin_annot_notmat/ConvEncoderUMAP',
                    'spect_params': {'fft_size': 512,
                                    'step_size': 32,
                                    'transform_type': 'log_spect_plus_one'},
                    'test_dur': 0.2
                },
        ]
    )
    def test_neither_audio_nor_spect_format_raises(
        self, config_dict
    ):
        """test that a config without either an audio or a spect format raises a ValueError"""
        # need to do this set-up so we don't mask one error with another
        config_dict['spect_params'] = vak.config.SpectParamsConfig(**config_dict['spect_params'])

        with pytest.raises(ValueError):
            prep_config = vak.config.PrepConfig(**config_dict)

    @pytest.mark.parametrize(
            'config_dict',
            [
                {
                    'annot_format': 'notmat',
                    'audio_format': 'cbin',
                    'data_dir': './tests/data_for_tests/source/audio_cbin_annot_notmat/gy6or6/032412',
                    'dataset_type': 'parametric umap',
                    'input_type': 'spect',
                    'labelset': 'iabcdefghjk',
                    'output_dir': './tests/data_for_tests/generated/prep/eval/audio_cbin_annot_notmat/ConvEncoderUMAP',
                    'spect_params': {'fft_size': 512,
                                    'step_size': 32,
                                    'transform_type': 'log_spect_plus_one'},
                    'test_dur': 0.2
                },
                {
                    'annot_format': 'notmat',
                    'audio_format': 'cbin',
                    'data_dir': './tests/data_for_tests/source/audio_cbin_annot_notmat/gy6or6/032312',
                    'dataset_type': 'frame classification',
                    'input_type': 'spect',
                    'labelset': 'iabcdefghjk',
                    'output_dir': './tests/data_for_tests/generated/prep/train/audio_cbin_annot_notmat/TweetyNet',
                    'spect_params': {'fft_size': 512,
                                    'freq_cutoffs': [500, 10000],
                                    'step_size': 64,
                                    'thresh': 6.25,
                                    'transform_type': 'log_spect'},
                    'test_dur': 30,
                    'train_dur': 50,
                    'val_dur': 15
                },
            ]
    )
    def test_from_config_dict(self, config_dict):
        # we have to make a copy since `from_config_dict` mutates the dict
        config_dict_copy = copy.deepcopy(config_dict)

        prep_config = vak.config.prep.PrepConfig.from_config_dict(config_dict_copy)

        assert isinstance(prep_config, vak.config.prep.PrepConfig)
        for key, val in config_dict.items():
            assert hasattr(prep_config, key)
            if key == 'data_dir' or key == 'output_dir':
                assert getattr(prep_config, key) == vak.common.converters.expanded_user_path(val)
            elif key == 'labelset':
                assert getattr(prep_config, key) == vak.common.converters.labelset_to_set(val)
            elif key == 'spect_params':
                assert getattr(prep_config, key) == vak.config.SpectParamsConfig(**val)
            else:
                assert getattr(prep_config, key) == val

    def test_from_config_dict_real_config(
            self, a_generated_config_dict
    ):
        prep_config = vak.config.prep.PrepConfig.from_config_dict(a_generated_config_dict['prep'])
        assert isinstance(prep_config, vak.config.prep.PrepConfig)



import pytest

import vak.config.trainer


class TestTrainerConfig:

    @pytest.mark.parametrize(
            'config_dict',
            [
                {
                    'accelerator': 'cpu',
                },
                {
                    'accelerator': 'gpu',
                    'devices': [0],
                },
                {
                    'accelerator': 'gpu',
                    'devices': [1],
                },
                {
                    'accelerator': 'cpu',
                    'devices': 1,
                },
                {
                    'devices': 1,
                },
            ]
    )
    def test_init(self, config_dict):
        trainer_config = vak.config.trainer.TrainerConfig(**config_dict)

        assert isinstance(trainer_config, vak.config.trainer.TrainerConfig)
        if 'accelerator' in config_dict:
            assert getattr(trainer_config, 'accelerator') == config_dict['accelerator']
        else:
            # TODO: mock `accelerator.get_default` here, return either 'cpu' or 'gpu'
            assert getattr(trainer_config, 'accelerator') == vak.common.accelerator.get_default()
        if 'devices' in config_dict:
            assert getattr(trainer_config, 'devices') == config_dict['devices']
        else:
            if 'accelerator' == 'cpu':
                assert getattr(trainer_config, 'devices') == 1
            elif 'accelerator' in ('gpu', 'tpu', 'ipu'):
                assert getattr(trainer_config, 'devices') == [0]

    @pytest.mark.parametrize(
            'config_dict, expected_exception',
            [
                # 'accelerator' can't be 'auto', breaks train/eval/prep/learncurve functions
                (
                    {
                        'accelerator': 'auto',
                    },
                    ValueError,
                ),
                # throws a device because parallel across GPUs won't work right now
                (
                    {
                        'accelerator': 'gpu',
                        'devices': [0, 1],
                    },
                    ValueError,
                ),
                # 'devices' can't be -1, breaks train/eval/prep/learncurve functions
                (
                    {
                        'accelerator': 'gpu',
                        'devices': -1,
                    },
                    ValueError,
                ),
                (
                    {
                        'accelerator': 'gpu',
                        'devices': 'auto',
                    },
                    ValueError,
                ),
                # when accelerator is CPU, devices must be int gt 0
                (
                    {
                        'accelerator': 'cpu',
                        'devices': -1,
                    },
                    ValueError,
                )
            ]
    )
    def test_init_raises(self, config_dict, expected_exception):
        with pytest.raises(expected_exception):
            vak.config.trainer.TrainerConfig(**config_dict)

    @pytest.mark.parametrize(
            'config_dict',
            [
                {
                    'accelerator': 'cpu',
                },
                {
                    'accelerator': 'gpu',
                },
                {
                    'accelerator': 'gpu',
                    'devices': [0],
                },
                {
                    'accelerator': 'gpu',
                    'devices': [1],
                },
                {
                    'accelerator': 'cpu',
                    'devices': 1,
                },
                {
                    'devices': 1,
                },
            ]
    )
    def test_asdict(self, config_dict):
        trainer_config = vak.config.trainer.TrainerConfig(**config_dict)

        trainer_config_asdict = trainer_config.asdict()

        assert isinstance(trainer_config_asdict, dict)
        if 'accelerator' in config_dict:
            assert trainer_config_asdict['accelerator'] == config_dict['accelerator']
        else:
            # TODO: mock `accelerator.get_default` here, return either 'cpu' or 'gpu'
            assert trainer_config_asdict['accelerator'] == vak.common.accelerator.get_default()
        if 'devices' in config_dict:
            assert trainer_config_asdict['devices'] == config_dict['devices']
        else:
            if config_dict["accelerator"] == 'cpu':
                assert trainer_config_asdict['devices'] == 1
            elif config_dict["accelerator"] == 'gpu':
                assert trainer_config_asdict['devices'] == [0]



import lightning
import torch

import vak.models.registry


# ---- mock networks ---------------------------------------------------------------------------------------------------
class MockNetwork(torch.nn.Module):
    """Network used just to test vak.models.base.Model"""
    def __init__(self, n_classes=10):
        super().__init__()
        # set these attributes so we can easily test config works
        self.n_classes = n_classes
        self.layers = torch.nn.Sequential(
            torch.nn.Linear(100, 50),
            torch.nn.Linear(50, 25),
            torch.nn.Linear(25, n_classes),
        )

    def forward(self, x):
        return self.layers(x)


class MockEncoder(torch.nn.Module):
    """Network used for testing.

    This network is put into a
    ``dict`` with ``MockDecoder`` to test
    that specifying ``network`` as a ``dict`` works.
    """
    def __init__(self, input_size=100, hidden_size=50):
        super().__init__()
        # set these attributes so we can easily test config works
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.layers = torch.nn.Sequential(
            torch.nn.Linear(input_size, hidden_size),
            torch.nn.Linear(hidden_size, hidden_size)
        )

    def forward(self, x):
        return self.layers(x)


class MockDecoder(torch.nn.Module):
    """Network used just to test vak.models.base.Model.
    Unlike ``MockNetwork``, this network will be put into a
    ``dict`` with ``MockDecoder`` to test
    that specifying ``network`` as a ``dict`` works.
    """
    def __init__(self, hidden_size=50, output_size=10):
        super().__init__()
        # set these attributes so we can easily test config works
        self.hidden_size = hidden_size
        self.output_size = output_size
        self.layers = torch.nn.Sequential(
            torch.nn.Linear(hidden_size, hidden_size),
            torch.nn.Linear(hidden_size, output_size)
        )

    def forward(self, x):
        return self.layers(x)


# ---- mock metrics ----------------------------------------------------------------------------------------------------
class MockAcc:
    """Mock metric used for testing"""
    def __init__(self, average='macro'):
        self.average = average

    def __call__(self, y: torch.Tensor, y_pred: torch.Tensor):
        sample_acc = y == y_pred
        if self.average == 'macro':
            return sample_acc.mean()
        elif self.average == 'micro':
            return NotImplemented


# ---- mock model families ---------------------------------------------------------------------------------------------
class UnregisteredMockModelFamily(lightning.LightningModule):
    """A model family defined only for tests.
    Used to test :func:`vak.models.registry.model_family`.
    """
    def __init__(self, network, optimizer, loss, metrics):
        super().__init__()
        self.network=network
        self.loss=loss
        self.optimizer=optimizer
        self.metrics=metrics

    def training_step(self, *args, **kwargs):
        pass

    def validation_step(self, *args, **kwargs):
        pass


# Make a "copy" of UnregisteredModelFamily that we *do* register
# so we can use it to test `vak.models.decorator.model` and other functions
# that require a registered ModelFamily.
# Used when testing :func:`vak.models.decorator.model` -- we need a model in the registry to test
# and we don't want to have to deal with the idiosyncrasies of actual model families
class MockModelFamily(lightning.LightningModule):
    """A model family defined only for tests.
    Used to test :func:`vak.models.registry.model_family`.
    """
    def __init__(self, network, optimizer, loss, metrics):
        super().__init__()
        self.network=network
        self.loss=loss
        self.optimizer=optimizer
        self.metrics=metrics

    def training_step(self, *args, **kwargs):
        pass

    def validation_step(self, *args, **kwargs):
        pass

vak.models.registry.model_family(MockModelFamily)


# ---- mock models -----------------------------------------------------------------------------------------------------
class MockModel:
    """Model definition used for testing :func:`vak.models.decorator.model`"""
    network = MockNetwork
    loss = torch.nn.CrossEntropyLoss
    optimizer = torch.optim.SGD
    metrics = {'acc': MockAcc}
    default_config = {
        'optimizer': {'lr': 0.003}
    }


class MockEncoderDecoderModel:
    """Model definition used for testing :func:`vak.models.decorator.model`.
    Specifically tests that `network` works with a ``dict``"""
    network = {'MockEncoder': MockEncoder, 'MockDecoder': MockDecoder}
    loss = torch.nn.TripletMarginWithDistanceLoss
    optimizer = torch.optim.Adam
    metrics = {
        'loss': torch.nn.TripletMarginWithDistanceLoss
    }
    default_config = {
        'optimizer': {'lr': 0.003}
    }


# pytest.mark.parametrize vals for test_init_with_definition
class OtherNetwork(torch.nn.Module):
    pass


class OtherOptimizer(torch.optim.Optimizer):
    pass


def other_loss_func(targets, y_pred):
    return


other_metrics_dict = {
    'other_metric': lambda x: x
}



import logging
import pathlib

import pandas as pd

from ...common.converters import expanded_user_path, labelset_to_set
from .. import constants
from ..audio_dataset import prep_audio_dataset
from ..spectrogram_dataset.prep import prep_spectrogram_dataset

logger = logging.getLogger(__name__)


def get_or_make_source_files(
    data_dir: str | pathlib.Path,
    input_type: str,
    audio_format: str | None = None,
    spect_format: str | None = None,
    spect_params: dict | None = None,
    spect_output_dir: str | pathlib.Path | None = None,
    annot_format: str | None = None,
    annot_file: str | pathlib.Path | None = None,
    labelset: set | None = None,
    audio_dask_bag_kwargs: dict | None = None,
) -> pd.DataFrame:
    """Get source files for a dataset, or make them.

    Gets either audio or spectrogram files from ``data dir``,
    possibly paired with annotation files.

    If ``input_type`` is ``'audio'``, then this function will look
    for files with the extension for ``audio_format`` in ``data_dir``.
    If ``input_type`` is ``'spectrogram'``, and ``spect_format`` is specified,
    then this function will look for files with the extension for that format
    in ``data_dir``. If ``input_type`` is spectrogram,
    and ``audio_format`` is specified,
    this function will look for audio files with that extension
    and then generate spectrograms for them using ``spect_params``.
    If an ``annot_format`` is specified, this function will additionally
    look for annotation files for the audio or spectrogram files.
    If all annotations are in a single file, this can be specified
    with the ``annot_file`` parameter, and that will be used instead
    of looking for other annotation files.

    Parameters
    ----------
    data_dir : str, Path
        Path to directory with files from which to make dataset.
    input_type : str
        The type of input to the neural network model.
        One of {'audio', 'spect'}.
    audio_format : str
        Format of audio files. One of {'wav', 'cbin'}.
        Default is ``None``, but either ``audio_format`` or ``spect_format``
        must be specified.
    spect_format : str
        Format of files containing spectrograms as 2-d matrices. One of {'mat', 'npz'}.
        Default is None, but either audio_format or spect_format must be specified.
    spect_params : dict, vak.config.SpectParams
        Parameters for creating spectrograms. Default is ``None``.
    spect_output_dir : str
        Path to location where spectrogram files should be saved.
        Default is None. If ``input_type`` is ``'spect'``,
        then ``spect_output_dir`` defaults to ``data_dir``.
    annot_format : str
        Format of annotations. Any format that can be used with the
        :module:`crowsetta` library is valid. Default is ``None``.
    annot_file : str
        Path to a single annotation file. Default is ``None``.
        Used when a single file contains annotates multiple audio
        or spectrogram files.
    audio_dask_bag_kwargs : dict
        Keyword arguments used when calling :func:`dask.bag.from_sequence`
        inside :func:`vak.io.audio`, where it is used to parallelize
        the conversion of audio files into spectrograms.
        Option should be specified in config.toml file as an inline table,
        e.g., ``audio_dask_bag_kwargs = { npartitions = 20 }``.
        Allows for finer-grained control
        when needed to process files of different sizes.
    labelset : str, list, set
        Set of unique labels for vocalizations. Strings or integers.
        Default is ``None``. If not ``None``, then files will be skipped
        where the associated annotation
        contains labels not found in ``labelset``.
        ``labelset`` is converted to a Python ``set`` using
        :func:`vak.converters.labelset_to_set`.
        See help for that function for details on how to specify ``labelset``.

    Returns
    -------
    source_files_df : pandas.DataFrame
        Source files that will become the dataset,
        represented as a pandas.DataFrame.
        Each row corresponds to one sample in the dataset,
        either an audio file or spectrogram file,
        possibly paired with annotations.
    """
    if input_type not in constants.INPUT_TYPES:
        raise ValueError(
            f"``input_type`` must be one of: {constants.INPUT_TYPES}\n"
            f"Value for ``input_type`` was: {input_type}"
        )

    if input_type == "audio" and spect_format is not None:
        raise ValueError(
            f"Input type was 'audio' but a ``spect_format`` was specified: '{spect_format}'. "
            f"Please specify ``audio_format`` only."
        )

    if input_type == "audio" and audio_format is None:
        raise ValueError(
            "Input type was 'audio' but no ``audio_format`` was specified. "
        )

    if audio_format is None and spect_format is None:
        raise ValueError(
            "Must specify either ``audio_format`` or ``spect_format``"
        )

    if audio_format and spect_format:
        raise ValueError(
            "Cannot specify both ``audio_format`` and ``spect_format``, "
            "unclear whether to create spectrograms from audio files or "
            "use already-generated spectrograms from array files"
        )

    if labelset is not None:
        labelset = labelset_to_set(labelset)

    data_dir = expanded_user_path(data_dir)
    if not data_dir.is_dir():
        raise NotADirectoryError(
            f"Path specified for ``data_dir`` not found: {data_dir}"
        )

    if annot_file is not None:
        annot_file = expanded_user_path(annot_file)
        if not annot_file.exists():
            raise FileNotFoundError(
                f"Path specified for ``annot_file`` not found: {annot_file}"
            )

    if input_type == "spect":
        source_files_df = prep_spectrogram_dataset(
            data_dir,
            annot_format,
            labelset,
            annot_file,
            audio_format,
            spect_format,
            spect_params,
            spect_output_dir,
            audio_dask_bag_kwargs,
        )
        if source_files_df.empty:
            raise ValueError(
                "Calling `vak.prep.spectrogram_dataset.prep_spectrogram_dataset` "
                "with arguments passed to `vak.prep.prep_frame_classification_dataset` "
                "returned an empty dataframe.\n"
                "Please double-check arguments to `prep_frame_classification_dataset` function."
            )

    elif input_type == "audio":
        source_files_df = prep_audio_dataset(
            audio_format,
            data_dir,
            annot_format,
            labelset,
        )
        if source_files_df.empty:
            raise ValueError(
                "Calling `vak.prep.audio_dataset.prep_audio_dataset` "
                "with arguments passed to `vak.prep.prep_frame_classification_dataset` "
                "returned an empty dataframe.\n"
                "Please double-check arguments to `prep_frame_classification_dataset` function."
            )

    return source_files_df



def get_train_dur_replicate_subset_name(
    train_dur: int, replicate_num: int
) -> str:
    """Get name of a training set subset for a learning curve,
    for a specified training duration and replicate number.

    Used when preparing the training set subsets for a learning curve,
    and when training models to generate the results for the curve.
    """
    return f"train-dur-{float(train_dur)}-replicate-{int(replicate_num)}"



"""fixtures relating to annotation files"""
import crowsetta
import pytest
import tomlkit


from .config import GENERATED_TEST_CONFIGS_ROOT
from .test_data import SOURCE_TEST_DATA_ROOT



ANNOT_FILE_YARDEN = SOURCE_TEST_DATA_ROOT.joinpath(
        "spect_mat_annot_yarden", "llb3", "llb3_annot_subset.mat"
    )


@pytest.fixture
def annot_file_yarden():
    return ANNOT_FILE_YARDEN


scribe_yarden = crowsetta.Transcriber(format="yarden")
ANNOT_LIST_YARDEN = scribe_yarden.from_file(ANNOT_FILE_YARDEN).to_annot()


@pytest.fixture
def annot_list_yarden():
    return ANNOT_LIST_YARDEN


LABELSET_YARDEN = [
    str(an_int)
    for an_int in [1, 2, 3, 6, 7, 8, 9, 10, 11, 12, 13, 14, 17, 18, 19]
]


@pytest.fixture
def labelset_yarden():
    """labelset as it would be loaded from a toml file

    don't return a set because we need to use this to test functions that convert it to a set.
    We also don't use a config for this since it's entered there as a "label string"
    """
    return LABELSET_YARDEN


ANNOT_DIR_NOTMAT = SOURCE_TEST_DATA_ROOT.joinpath("audio_cbin_annot_notmat", "gy6or6", "032312")


@pytest.fixture
def annot_dir_notmat():
    return ANNOT_DIR_NOTMAT


ANNOT_FILES_NOTMAT = sorted(ANNOT_DIR_NOTMAT.glob("*.not.mat"))


@pytest.fixture
def annot_files_notmat():
    return ANNOT_FILES_NOTMAT


scribe_notmat = crowsetta.Transcriber(format="notmat")
ANNOT_LIST_NOTMAT = [scribe_notmat.from_file(notmat).to_annot() for notmat in ANNOT_FILES_NOTMAT]


@pytest.fixture
def annot_list_notmat():
    return ANNOT_LIST_NOTMAT



a_train_notmat_config = sorted(
    GENERATED_TEST_CONFIGS_ROOT.glob("*train*notmat*toml")
)[0]  # get first config.toml from glob list
# doesn't really matter which config, they all have labelset
with a_train_notmat_config.open("r") as fp:
    a_train_notmat_toml = tomlkit.load(fp)
LABELSET_NOTMAT = list(str(a_train_notmat_toml["vak"]["prep"]["labelset"]))


@pytest.fixture
def labelset_notmat(generated_test_configs_root):
    """labelset as it would be loaded from a toml file

    don't return a set because we need to use this to test functions that convert it to a set"""
    return LABELSET_NOTMAT


@pytest.fixture
def specific_annot_list(annot_list_notmat,
                        annot_list_yarden):
    """factory fixture, returns a function that
    returns a fixture containing a list of Annotation objects,
    given a specified annotation format

    so that unit tests can be parameterized with annotation format names
    """
    FORMAT_ANNOT_LIST_FIXTURE_MAP = {
        "notmat": annot_list_notmat,
        "yarden": annot_list_yarden,
    }

    def _annot_list_factory(format):
        return FORMAT_ANNOT_LIST_FIXTURE_MAP[format]

    return _annot_list_factory


@pytest.fixture
def specific_labelset(labelset_yarden, labelset_notmat):
    def _specific_labelset(annot_format):
        if annot_format == "yarden":
            return labelset_yarden
        elif annot_format == "notmat":
            return labelset_notmat
        else:
            raise ValueError(f"invalid annot_format: {annot_format}")

    return _specific_labelset


ANNOT_FILES_WITH_NO_SEGMENTS_DIR = SOURCE_TEST_DATA_ROOT / 'audio_cbin_annot_notmat' / 'gy6or6-annotated-with-no-labels'
ANNOTATED_FILES_NO_SEGMENTS = sorted(ANNOT_FILES_WITH_NO_SEGMENTS_DIR.glob('*.cbin'))
ANNOT_FILES_WITH_NO_SEGMENTS = sorted(ANNOT_FILES_WITH_NO_SEGMENTS_DIR.glob('*.not.mat'))
ANNOTATED_ANNOT_NO_SEGMENTS_TUPLES = list(
    zip(ANNOTATED_FILES_NO_SEGMENTS, ANNOT_FILES_WITH_NO_SEGMENTS)
)


@pytest.fixture(params=ANNOTATED_ANNOT_NO_SEGMENTS_TUPLES)
def annotated_annot_no_segments(request):
    """Tuple of (annotated (audio) path, annotation path),
    where the annotation file has no annotated segments in it.
    Used to test edge case for `has_unlabeled`,
    see https://github.com/vocalpy/vak/issues/378
    """

    return request.param



"""Assign samples in a dataset to splits.

Given a set of source files represented by a dataframe,
assign each sample (row) to a split.

Helper function called by :func:`vak.prep.frame_classification.prep_frame_classification_dataset`.
"""

from __future__ import annotations

import logging
import pathlib

import pandas as pd

from .. import dataset_df_helper, split

logger = logging.getLogger(__name__)


def assign_samples_to_splits(
    purpose: str,
    dataset_df: pd.DataFrame,
    dataset_path: str | pathlib.Path,
    train_dur: float | None = None,
    val_dur: float | None = None,
    test_dur: float | None = None,
    labelset: set | None = None,
) -> pd.DataFrame:
    """Assign samples in a dataset to splits.

    Given a set of source files represented by a dataframe,
    assign each sample (row) to a split.

    Helper function called by :func:`vak.prep.frame_classification.prep_frame_classification_dataset`.

    If no durations are specified for splits,
    or the purpose is either `'eval'` or `'predict'`,
    then all rows in the dataframe
    will be assigned to ``purpose``.

    Parameters
    ----------
    purpose : str
        Purpose of the dataset.
        One of {'train', 'eval', 'predict', 'learncurve'}.
        These correspond to commands of the vak command-line interface.
    train_dur : float
        Total duration of training set, in seconds.
        When creating a learning curve,
        training subsets of shorter duration
        will be drawn from this set. Default is None.
    val_dur : float
        Total duration of validation set, in seconds.
        Default is None.
    test_dur : float
        Total duration of test set, in seconds.
        Default is None.
    dataset_df : pandas.DataFrame
        That represents a dataset.
    dataset_path : pathlib.Path
        Path to csv saved from ``dataset_df``.
    labelset : str, list, set
        Set of unique labels for vocalizations. Strings or integers.
        Default is ``None``. If not ``None``, then files will be skipped
        where the associated annotation
        contains labels not found in ``labelset``.
        ``labelset`` is converted to a Python ``set`` using
        :func:`vak.converters.labelset_to_set`.
        See help for that function for details on how to specify ``labelset``.

    Returns
    -------
    dataset_df : pandas.DataFrame
        The same ``dataset_df`` with a `'split'` column added,
        where each element in that column assigns the corresponding
        row to one of the splits in the dataset.
    """

    # ---- (possibly) split into train / val / test sets ---------------------------------------------
    # catch case where user specified duration for just training set, raise a helpful error instead of failing silently
    if (purpose == "train" or purpose == "learncurve") and (
        (train_dur is not None and train_dur > 0)
        and (val_dur is None or val_dur == 0)
        and (test_dur is None or val_dur == 0)
    ):
        raise ValueError(
            "A duration specified for just training set, but prep function does not currently support creating a "
            "single split of a specified duration. Either remove the train_dur option from the prep section and "
            "rerun, in which case all data will be included in the training set, or specify values greater than "
            "zero for test_dur (and val_dur, if a validation set will be used)"
        )

    if all(
        [dur is None for dur in (train_dur, val_dur, test_dur)]
    ) or purpose in (
        "eval",
        "predict",
    ):
        # then we're not going to split
        logger.info("Will not split dataset.")
        do_split = False
    else:
        if val_dur is not None and train_dur is None and test_dur is None:
            raise ValueError(
                "cannot specify only val_dur, unclear how to split dataset into training and test sets"
            )
        else:
            logger.info("Will split dataset.")
            do_split = True

    if do_split:
        dataset_df = split.frame_classification_dataframe(
            dataset_df,
            dataset_path,
            labelset=labelset,
            train_dur=train_dur,
            val_dur=val_dur,
            test_dur=test_dur,
        )

    elif (
        do_split is False
    ):  # add a split column, but assign everything to the same 'split'
        # ideally we would just say split=purpose in call to add_split_col, but
        # we have to special case, because "eval" looks for a 'test' split (not an "eval" split)
        if purpose == "eval":
            split_name = (
                "test"  # 'split_name' to avoid name clash with split package
            )
        elif purpose == "predict":
            split_name = "predict"

        dataset_df = dataset_df_helper.add_split_col(
            dataset_df, split=split_name
        )

    return dataset_df



"""Validators for frame classification datasets"""

from __future__ import annotations

import pandas as pd


def validate_and_get_frame_dur(
    dataset_df: pd.DataFrame, input_type: str
) -> float:
    """Validate that there is a single, unique value for the
    frame duration for all samples (audio signals / spectrograms)
    in a dataset. If so, return that value.

    The dataset is represented as a pandas DataFrame.

    Parameters
    ----------
    dataset_df : pandas.Dataframe
        A pandas.DataFrame created by
        :func:`vak.prep.spectrogram_dataset.prep_spectrogram_dataset`
        or :func:`vak.prep.audio_dataset.prep_audio_dataset`.
    input_type : str
        The type of input to the neural network model.
        One of {'audio', 'spect'}.

    Returns
    -------
    frame_dur : float
        The duration of a time bin in seconds
        for all spectrograms in the dataset.
    """
    from .. import constants  # avoid circular import

    if input_type not in constants.INPUT_TYPES:
        raise ValueError(
            f"``input_type`` must be one of: {constants.INPUT_TYPES}\n"
            f"Value for ``input_type`` was: {input_type}"
        )

    # TODO: handle possible KeyError here?
    if input_type == "audio":
        frame_dur = dataset_df["sample_dur"].unique()
    elif input_type == "spect":
        frame_dur = dataset_df["timebin_dur"].unique()

    if len(frame_dur) > 1:
        raise ValueError(
            f"Found more than one frame duration in dataset: {frame_dur}"
        )

    frame_dur = frame_dur.item()

    return frame_dur



"""Class representing the model table of a toml configuration file."""

from __future__ import annotations

from attrs import asdict, define, field
from attrs.validators import instance_of

from .. import models

MODEL_TABLES = [
    "network",
    "optimizer",
    "loss",
    "metrics",
]


@define
class ModelConfig:
    """Class representing the model table of a toml configuration file.

    Attributes
    ----------
    name : str
    network : dict
        Keyword arguments for the network class,
        or a :class:`dict` of ``dict``s mapping
        network names to keyword arguments.
    optimizer: dict
        Keyword arguments for the optimizer class.
    loss : dict
        Keyword arguments for the class representing the loss function.
    metrics: dict
        A :class:`dict` of ``dict``s mapping
        metric names to keyword arguments.
    """

    name: str
    network: dict = field(validator=instance_of(dict))
    optimizer: dict = field(validator=instance_of(dict))
    loss: dict = field(validator=instance_of(dict))
    metrics: dict = field(validator=instance_of(dict))

    @classmethod
    def from_config_dict(cls, config_dict: dict):
        """Return :class:`ModelConfig` instance from a :class:`dict`.

        The :class:`dict` passed in should be the one found
        by loading a valid configuration toml file with
        :func:`vak.config.parse.from_toml_path`,
        and then using a top-level table key,
        followed by key ``'model'``.
        E.g., ``config_dict['train']['model']` or
        ``config_dict['predict']['model']``.

        Examples
        --------
        config_dict = vak.config.parse.from_toml_path(toml_path)
        model_config = vak.config.Model.from_config_dict(config_dict['train'])
        """
        model_name = list(config_dict.keys())
        if len(model_name) == 0:
            raise ValueError(
                "Did not find a single key in `config_dict` corresponding to model name. "
                f"Instead found no keys. Config dict:\n{config_dict}\n"
                "A configuration file should specify a single model per top-level table."
            )
        if len(model_name) > 1:
            raise ValueError(
                "Did not find a single key in `config_dict` corresponding to model name. "
                f"Instead found multiple keys: {model_name}.\nConfig dict:\n{config_dict}.\n"
                "A configuration file should specify a single model per top-level table."
            )
        model_name = model_name[0]
        MODEL_NAMES = list(models.registry.MODEL_NAMES)
        if model_name not in MODEL_NAMES:
            raise ValueError(
                f"Model name not found in registry: {model_name}\n"
                f"Model names in registry:\n{MODEL_NAMES}"
            )

        # NOTE: we are getting model_config here
        model_config = config_dict[model_name]
        if not all(key in MODEL_TABLES for key in model_config.keys()):
            invalid_keys = (
                key for key in model_config.keys() if key not in MODEL_TABLES
            )
            raise ValueError(
                f"The following sub-tables in the model config are not valid: {invalid_keys}\n"
                f"Valid sub-table names are: {MODEL_TABLES}"
            )
        # for any tables not specified, default to empty dict so we can still use ``**`` operator on it
        for model_table in MODEL_TABLES:
            if model_table not in model_config:
                model_config[model_table] = {}
        return cls(name=model_name, **model_config)

    def asdict(self):
        """Convert this :class:`ModelConfig` instance
        to a :class:`dict` that can be passed
        into functions that take a ``model_config`` argument,
        like :func:`vak.train` and :func:`vak.predict`.
        """
        return asdict(self)



from . import conv_encoder, ed_tcn, tweetynet
from .conv_encoder import ConvEncoder
from .ed_tcn import ED_TCN
from .tweetynet import TweetyNet

__all__ = [
    "conv_encoder",
    "ConvEncoder",
    "ed_tcn",
    "ED_TCN",
    "tweetynet",
    "TweetyNet",
]



"""fixtures relating to .toml configuration files"""
import json
import shutil

import pytest
import tomlkit

from .test_data import GENERATED_TEST_DATA_ROOT, TEST_DATA_ROOT


TEST_CONFIGS_ROOT = TEST_DATA_ROOT.joinpath("configs")


@pytest.fixture
def test_configs_root():
    """Path that points to data_for_tests/configs

    Two types of config files in this directory:
    1) those used by the tests/scripts/generate_data_for_tests.py script.
       Will be listed in configs.json. See ``specific_config_toml_path`` fixture below
       for details about types of configs.
    2) those used by tests that are static, e.g., ``invalid_table_config.toml``

    This fixture facilitates access to type (2), e.g. in test_config/test_parse
    """
    return TEST_CONFIGS_ROOT


@pytest.fixture
def list_of_schematized_configs(test_configs_root):
    """returns list of configuration files,
    schematized with attributes
    so that fixtures and unit tests can specify those attributes
    to find the filename of a specific configuration file,
    that can then be used to get that file.

    Each element in the list is a dict with the following keys:
    `filename`, `config_type`, `audio_format`, `spect_format`, `annot_format`
    These keys define the schema for config files.

    For example, here is the first one:
    {
      "filename": "test_eval_audio_cbin_annot_notmat.toml",
      "config_type": "eval",
      "audio_format": "cbin",
      "spect_format": null,
      "annot_format": "notmat"
    }

    The ``specific_config_toml_path`` factory fixture returns a function that
    itself return a configuration ``filename``, when provided values for
    all of the other keys.
    """
    with test_configs_root.joinpath("configs.json").open("r") as fp:
        return json.load(fp)["config_metadata"]


@pytest.fixture
def config_that_doesnt_exist(tmp_path):
    return tmp_path / "config_that_doesnt_exist.toml"


@pytest.fixture
def invalid_table_config_path(test_configs_root):
    return test_configs_root.joinpath("invalid_table_config.toml")


@pytest.fixture
def invalid_key_config_path(test_configs_root):
    return test_configs_root.joinpath("invalid_key_config.toml")


@pytest.fixture
def invalid_train_and_learncurve_config_toml(test_configs_root):
    return test_configs_root.joinpath("invalid_train_and_learncurve_config.toml")


GENERATED_TEST_CONFIGS_ROOT = GENERATED_TEST_DATA_ROOT.joinpath("configs")


@pytest.fixture
def generated_test_configs_root():
    return GENERATED_TEST_CONFIGS_ROOT


@pytest.fixture
def specific_config_toml_path(generated_test_configs_root, list_of_schematized_configs, tmp_path):
    """returns a factory function
    that will return the path
    to a specific configuration file, determined by
    characteristics specified by the caller:
    `config_type`, `audio_format`, `spect_format`, `annot_format`

    The factory function actually returns a copy,
    that will be copied into ``tmp_path``,
    so the original remains unmodified.

    If ``root_results_dir`` argument is specified
    when calling the factory function,
    it will convert the value for that key in the table
    corresponding to ``config_type`` to the value
    specified for ``root_results_dir``.
    This makes it possible to dynamically change the ``root_results_dir``
    e.g. to the ``tmp_path`` fixture used by unit tests
    """

    def _specific_config(
        config_type,
        model,
        annot_format,
        audio_format=None,
        spect_format=None,
        keys_to_change=None,
    ):
        """returns path to a specific configuration file,
        determined by characteristics specified by the caller:
        `config_type`, `audio_format`, `spect_format`, `annot_format`

        Parameters
        ----------
        config_type : str
            corresponding to a `vak` cli command
        annot_format : str
            annotation format, recognized by ``crowsetta``
        audio_format : str
        spect_format : str
        keys_to_change : list, dict
            list of dicts with keys 'table', 'key', and 'value'.
            Can be a single dict, in which case only that key is changed.
            If the 'value' is set to 'DELETE-KEY',
            the key will be removed from the config.
            This can be used to test behavior when the key is not set.

        Returns
        -------
        config_path : pathlib.Path
            that points to temporary copy of specified config,
            with any keys changed as specified
        """
        original_config_path = None
        for schematized_config in list_of_schematized_configs:
            if all(
                [
                    schematized_config["config_type"] == config_type,
                    schematized_config["model"] == model,
                    schematized_config["annot_format"] == annot_format,
                    schematized_config["audio_format"] == audio_format,
                    schematized_config["spect_format"] == spect_format,
                ]
            ):
                original_config_path = generated_test_configs_root.joinpath(
                    schematized_config["filename"]
                )
                break

        if original_config_path is None:
            raise ValueError(
                f"did not find a specific config with `config_type`='{config_type}', "
                f"`model`={model}, `annot_format`='{annot_format}', "
                f"`audio_format`='{audio_format}', and `spect_type`='{spect_format}'."
            )
        config_copy_path = tmp_path.joinpath(original_config_path.name)
        config_copy_path = shutil.copy(src=original_config_path, dst=config_copy_path)

        if keys_to_change is not None:
            if isinstance(keys_to_change, dict):
                keys_to_change = [keys_to_change]
            elif isinstance(keys_to_change, list):
                pass
            else:
                raise TypeError(
                    f"invalid type for `keys_to_change`: {type(keys_to_change)}"
                )

            with config_copy_path.open("r") as fp:
                tomldoc = tomlkit.load(fp)

            for table_key_val_dict in keys_to_change:
                table_name = table_key_val_dict["table"]
                key = table_key_val_dict["key"]
                value = table_key_val_dict["value"]
                if isinstance(key, str):
                    if table_key_val_dict["value"] == 'DELETE-KEY':
                        # e.g., to test behavior of config without this key
                        del tomldoc["vak"][table_name][key]
                    else:
                        tomldoc["vak"][table_name][key] = value
                elif isinstance(key, list) and len(key) == 2 and all([isinstance(el, str) for el in key]):
                    # for the case where we need to access a sub-table
                    # right now this applies mainly to ["vak"][table]["dataset"]["path"]
                    # if we end up having to access more / deeper then we'll need something more general
                    if table_key_val_dict["value"] == 'DELETE-KEY':
                        # e.g., to test behavior of config without this key
                        del tomldoc["vak"][table_name][key[0]][key[1]]
                    else:
                        tomldoc["vak"][table_name][key[0]][key[1]] = value
                else:
                    raise ValueError(
                        f"Unexpected value for 'key' in `keys_to_change` dict: {key}.\n"
                        f"`keys_to_change` dict: {table_key_val_dict}"
                    )

            with config_copy_path.open("w") as fp:
                tomlkit.dump(tomldoc, fp)

        return config_copy_path

    return _specific_config


ALL_GENERATED_CONFIG_PATHS = sorted(GENERATED_TEST_CONFIGS_ROOT.glob("*toml"))


# ---- path to config files ----
@pytest.fixture(params=ALL_GENERATED_CONFIG_PATHS)
def a_generated_config_path(request):
    return request.param


def _tomlkit_to_popo(d):
    """Convert tomlkit to "popo" (Plain-Old Python Objects)

    From https://github.com/python-poetry/tomlkit/issues/43#issuecomment-660415820
    """
    try:
        result = getattr(d, "value")
    except AttributeError:
        result = d

    if isinstance(result, list):
        result = [_tomlkit_to_popo(x) for x in result]
    elif isinstance(result, dict):
        result = {
            _tomlkit_to_popo(key): _tomlkit_to_popo(val) for key, val in result.items()
        }
    elif isinstance(result, tomlkit.items.Integer):
        result = int(result)
    elif isinstance(result, tomlkit.items.Float):
        result = float(result)
    elif isinstance(result, tomlkit.items.String):
        result = str(result)
    elif isinstance(result, tomlkit.items.Bool):
        result = bool(result)

    return result


# ----  config dicts from paths ----
def _load_config_dict(toml_path):
    """Return config as dict, loaded from toml file.

    Used to test functions that parse config tables, taking these dicts as inputs.

    Note that we access the topmost table loaded from the toml: config_dict['vak']
    """
    with toml_path.open("r") as fp:
        config_dict = tomlkit.load(fp)
    return _tomlkit_to_popo(config_dict['vak'])


@pytest.fixture
def specific_config_toml(specific_config_toml_path):
    """returns a function that will return a dict
    containing parsed toml from a
    specific configuration file, determined by
    characteristics specified by the caller:
    `config_type`, `audio_format`, `spect_format`, `annot_format`
    """

    def _specific_config_toml(
        config_type,
        model,
        annot_format,
        audio_format=None,
        spect_format=None,
    ):
        config_path = specific_config_toml_path(
            config_type, model, annot_format, audio_format, spect_format
        )
        return _load_config_dict(config_path)

    return _specific_config_toml


@pytest.fixture(params=ALL_GENERATED_CONFIG_PATHS)
def a_generated_config_dict(request):
    # we remake dict every time this gets called
    # so that we're not returning a ``config_dict`` that was
    # already mutated by a `Config.from_config_dict` function,
    # e.g. the value for the 'spect_params' key gets mapped to a SpectParamsConfig
    # by PrepConfig.from_config_dict
    return _load_config_dict(request.param)


ALL_GENERATED_EVAL_CONFIG_PATHS = sorted(
    GENERATED_TEST_CONFIGS_ROOT.glob("*eval*toml")
)

ALL_GENERATED_LEARNCURVE_CONFIG_PATHS = sorted(
    GENERATED_TEST_CONFIGS_ROOT.glob("*learncurve*toml")
)

ALL_GENERATED_PREDICT_CONFIG_PATHS = sorted(
    GENERATED_TEST_CONFIGS_ROOT.glob("*predict*toml")
)

ALL_GENERATED_TRAIN_CONFIG_PATHS = sorted(
    GENERATED_TEST_CONFIGS_ROOT.glob("*train*toml")
)

# as above, we remake dict every time these fixutres get called
# so that we're not returning a ``config_dict`` that was
# already mutated by a `Config.from_config_dict` function,
# e.g. the value for the 'spect_params' key gets mapped to a SpectParamsConfig
# by PrepConfig.from_config_dict
@pytest.fixture(params=ALL_GENERATED_EVAL_CONFIG_PATHS)
def a_generated_eval_config_dict(request):
    return _load_config_dict(request.param)


@pytest.fixture(params=ALL_GENERATED_LEARNCURVE_CONFIG_PATHS)
def a_generated_learncurve_config_dict(request):
    return _load_config_dict(request.param)


@pytest.fixture(params=ALL_GENERATED_PREDICT_CONFIG_PATHS)
def a_generated_predict_config_dict(request):
    return _load_config_dict(request.param)


@pytest.fixture(params=ALL_GENERATED_TRAIN_CONFIG_PATHS)
def a_generated_train_config_dict(request):
    return _load_config_dict(request.param)


@pytest.fixture
def all_generated_learncurve_configs_toml(all_generated_learncurve_configs):
    return [_load_config_dict(config) for config in all_generated_learncurve_configs]


ALL_GENERATED_CONFIGS_TOML_PATH_PAIRS = list(zip(
    [_load_config_dict(config) for config in ALL_GENERATED_CONFIG_PATHS],
    ALL_GENERATED_CONFIG_PATHS,
))


# ---- config toml + path pairs ----
@pytest.fixture
def all_generated_configs_toml_path_pairs():
    """zip of tuple pairs: (dict, pathlib.Path)
    where ``Path`` is path to .toml config file and ``dict`` is
    the .toml config from that path
    loaded into a dict with the ``toml`` library
    """
    # we duplicate the constant above because we need to remake
    # the variables for each unit test. Otherwise tests that modify values
    # for config keys cause other tests to fail
    return zip(
        [_load_config_dict(config) for config in ALL_GENERATED_CONFIG_PATHS],
        ALL_GENERATED_CONFIG_PATHS
    )


@pytest.fixture
def configs_toml_path_pairs_by_model_factory(all_generated_configs_toml_path_pairs):
    """
    factory fixture that returns all generated configs and their paths for a specified model
    """

    def _wrapped(model,
                 table_name=None):
        out = []
        unzipped = list(all_generated_configs_toml_path_pairs)
        for config_toml, toml_path in unzipped:
            if toml_path.name.startswith(model):
                if table_name:
                    if table_name.lower() in toml_path.name:
                        out.append(
                            (config_toml, toml_path)
                        )
                else:
                    out.append(
                     (config_toml, toml_path)
                    )
        return out

    return _wrapped




import pytest

import vak
import vak.datapipes.parametric_umap


class TestDatapipe:
    @pytest.mark.parametrize(
        'config_type, model_name, audio_format, spect_format, annot_format, split, transform_kwargs',
        [
            ('train', 'ConvEncoderUMAP', 'cbin', None, 'notmat', 'train', {}),
        ]
    )
    def test_from_dataset_path(self, config_type, model_name, audio_format, spect_format, annot_format,
                               split, transform_kwargs, specific_config_toml_path):
        """Test we can get a :class:`vak.datapipes.parametric_umap.Datapipe` instance
        from the classmethod ``from_dataset_path``"""
        toml_path = specific_config_toml_path(config_type,
                                              model_name,
                                              audio_format=audio_format,
                                              spect_format=spect_format,
                                              annot_format=annot_format)
        cfg = vak.config.Config.from_toml_path(toml_path)
        cfg_command = getattr(cfg, config_type)

        dataset = vak.datapipes.parametric_umap.Datapipe.from_dataset_path(
            dataset_path=cfg_command.dataset.path,
            split=split,
        )
        assert isinstance(dataset, vak.datapipes.parametric_umap.Datapipe)



"""tests for vak.eval.frame_classification module"""
import pytest

import vak.config
import vak.common.constants
import vak.common.paths
import vak.eval.frame_classification


# written as separate function so we can re-use in tests/unit/test_cli/test_eval.py
def assert_eval_output_matches_expected(model_name, output_dir):
    eval_csv = sorted(output_dir.glob(f"eval_{model_name}*csv"))
    assert len(eval_csv) == 1


# -- we do eval with all possible configurations of post_tfm_kwargs
POST_TFM_KWARGS = [
    # default, will use ToLabels
    None,
    # no cleanup but uses ToLabelsWithPostprocessing
    {'majority_vote': False, 'min_segment_dur': None},
    # use ToLabelsWithPostprocessing with *just* majority_vote
    {'majority_vote': True, 'min_segment_dur': None},
    # use ToLabelsWithPostprocessing with *just* min_segment_dur
    {'majority_vote': False, 'min_segment_dur': 0.002},
    # use ToLabelsWithPostprocessing with majority_vote *and* min_segment_dur
    {'majority_vote': True, 'min_segment_dur': 0.002},
]


@pytest.fixture(params=POST_TFM_KWARGS)
def post_tfm_kwargs(request):
    return request.param


@pytest.mark.parametrize(
    "model_name, audio_format, spect_format, annot_format",
    [
        ("TweetyNet", "cbin", None, "notmat"),
    ],
)
def test_eval_frame_classification_model(
        model_name,
        audio_format,
        spect_format,
        annot_format,
        specific_config_toml_path,
        tmp_path,
        trainer_table,
        post_tfm_kwargs
):
    output_dir = tmp_path.joinpath(
        f"test_eval_{audio_format}_{spect_format}_{annot_format}"
    )
    output_dir.mkdir()

    keys_to_change = [
        {"table": "eval", "key": "output_dir", "value": str(output_dir)},
        {"table": "eval", "key": "trainer", "value": trainer_table},
    ]

    toml_path = specific_config_toml_path(
        config_type="eval",
        model=model_name,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)

    vak.eval.frame_classification.eval_frame_classification_model(
        model_config=cfg.eval.model.asdict(),
        dataset_config=cfg.eval.dataset.asdict(),
        trainer_config=cfg.eval.trainer.asdict(),
        checkpoint_path=cfg.eval.checkpoint_path,
        labelmap_path=cfg.eval.labelmap_path,
        output_dir=cfg.eval.output_dir,
        num_workers=cfg.eval.num_workers,
        frames_standardizer_path=cfg.eval.frames_standardizer_path,
        post_tfm_kwargs=post_tfm_kwargs,
    )

    assert_eval_output_matches_expected(cfg.eval.model.name, output_dir)


@pytest.mark.parametrize(
    'path_option_to_change',
    [
        {"table": "eval", "key": "checkpoint_path", "value": '/obviously/doesnt/exist/ckpt.pt'},
        {"table": "eval", "key": "labelmap_path", "value": '/obviously/doesnt/exist/labelmap.json'},
        {"table": "eval", "key": "frames_standardizer_path", "value": '/obviously/doesnt/exist/FramesStandardizer'},
    ]
)
def test_eval_frame_classification_model_raises_file_not_found(
    path_option_to_change,
        specific_config_toml_path,
    tmp_path,
    trainer_table
):
    """Test that core.eval raises FileNotFoundError
    when one of the following does not exist:
    checkpoint_path, labelmap_path, dataset_path, frames_standardizer_path
    """
    output_dir = tmp_path.joinpath(
        f"test_eval_cbin_notmat_invalid_dataset_path"
    )
    output_dir.mkdir()

    keys_to_change = [
        {"table": "eval", "key": "output_dir", "value": str(output_dir)},
        {"table": "eval", "key": "trainer", "value": trainer_table},
        path_option_to_change,
    ]

    toml_path = specific_config_toml_path(
        config_type="eval",
        model="TweetyNet",
        audio_format="cbin",
        annot_format="notmat",
        spect_format=None,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)
    with pytest.raises(FileNotFoundError):
        vak.eval.frame_classification.eval_frame_classification_model(
            model_config=cfg.eval.model.asdict(),
            dataset_config=cfg.eval.dataset.asdict(),
            trainer_config=cfg.eval.trainer.asdict(),
            checkpoint_path=cfg.eval.checkpoint_path,
            labelmap_path=cfg.eval.labelmap_path,
            output_dir=cfg.eval.output_dir,
            num_workers=cfg.eval.num_workers,
            frames_standardizer_path=cfg.eval.frames_standardizer_path,
        )


@pytest.mark.parametrize(
    'path_option_to_change',
    [
        {"table": "eval", "key": ["dataset","path"], "value": '/obviously/doesnt/exist/dataset-dir'},
        {"table": "eval", "key": "output_dir", "value": '/obviously/does/not/exist/output'},
    ]
)
def test_eval_frame_classification_model_raises_not_a_directory(
    path_option_to_change,
        specific_config_toml_path,
    trainer_table,
    tmp_path,
):
    """Test that core.eval raises NotADirectory
    when directories don't exist
    """
    keys_to_change = [
        path_option_to_change,
        {"table": "eval", "key": "trainer", "value": trainer_table},
    ]

    if path_option_to_change["key"] != "output_dir":
        # need to make sure output_dir *does* exist
        # so we don't detect spurious NotADirectoryError and assume test passes
        output_dir = tmp_path.joinpath(
            f"test_eval_raises_not_a_directory"
        )
        output_dir.mkdir()
        keys_to_change.append(
            {"table": "eval", "key": "output_dir", "value": str(output_dir)}
        )

    toml_path = specific_config_toml_path(
        config_type="eval",
        model="TweetyNet",
        audio_format="cbin",
        annot_format="notmat",
        spect_format=None,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)
    with pytest.raises(NotADirectoryError):
        vak.eval.frame_classification.eval_frame_classification_model(
            model_config=cfg.eval.model.asdict(),
            dataset_config=cfg.eval.dataset.asdict(),
            trainer_config=cfg.eval.trainer.asdict(),
            checkpoint_path=cfg.eval.checkpoint_path,
            labelmap_path=cfg.eval.labelmap_path,
            output_dir=cfg.eval.output_dir,
            num_workers=cfg.eval.num_workers,
            frames_standardizer_path=cfg.eval.frames_standardizer_path,
        )



from . import decorator, definition, factory, registry
from .convencoder_umap import ConvEncoderUMAP
from .decorator import model
from .ed_tcn import ED_TCN
from .factory import ModelFactory
from .frame_classification_model import FrameClassificationModel
from .get import get
from .parametric_umap_model import ParametricUMAPModel
from .registry import model_family
from .tweetynet import TweetyNet

__all__ = [
    "factory",
    "ConvEncoderUMAP",
    "decorator",
    "definition",
    "ED_TCN",
    "FrameClassificationModel",
    "get",
    "ModelFactory",
    "model",
    "model_family",
    "ParametricUMAPModel",
    "registry",
    "TweetyNet",
]



"""Function that generates new inferences from trained models in the frame classification family."""

from __future__ import annotations

import json
import logging
import os
import pathlib

import crowsetta
import joblib
import lightning
import numpy as np
import pandas as pd
import torch.utils.data
from attrs import define
from tqdm import tqdm

from .. import common, datapipes, datasets, models, transforms
from ..common import constants, files, validators
from ..datapipes.frame_classification import InferDatapipe

logger = logging.getLogger(__name__)


@define
class AnnotationDataFrame:
    """Data class that represents annotations
    for an audio file, in a :class:`pandas.DataFrame`.

    Used to save annotations that currently can't
    be saved with :mod:`crowsetta`, e.g. boundary times.
    """

    df: pd.DataFrame
    audio_path: str | pathlib.Path


def predict_with_frame_classification_model(
    model_config: dict,
    dataset_config: dict,
    trainer_config: dict,
    checkpoint_path: str | pathlib.Path,
    labelmap_path: str | pathlib.Path,
    num_workers: int = 2,
    timebins_key: str = "t",
    frames_standardizer_path: str | pathlib.Path | None = None,
    annot_csv_filename: str | None = None,
    output_dir: str | pathlib.Path | None = None,
    min_segment_dur: float | None = None,
    majority_vote: bool = False,
    save_net_outputs: bool = False,
    background_label: str = common.constants.DEFAULT_BACKGROUND_LABEL,
) -> None:
    """Make predictions on a dataset with a trained
    :class:`~vak.models.FrameClassificationModel`.

     Parameters
     ----------
    model_config : dict
        Model configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.ModelConfig.asdict`.
    dataset_config: dict
        Dataset configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.DatasetConfig.asdict`.
    trainer_config: dict
        Configuration for :class:`lightning.pytorch.Trainer`.
        Can be obtained by calling :meth:`vak.config.TrainerConfig.asdict`.
    checkpoint_path : str
        path to directory with checkpoint files saved by Torch, to reload model
    labelmap_path : str
        path to 'labelmap.json' file.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader. Default is 2.
    spect_key : str
        key for accessing spectrogram in files. Default is 's'.
    timebins_key : str
        key for accessing vector of time bins in files. Default is 't'.
    frames_standardizer_path : str
        path to a saved :class:`vak.transforms.FramesStandardizer` object used to standardize (normalize) frames.
        If spectrograms were normalized and this is not provided, will give
        incorrect results.
    annot_csv_filename : str
        name of .csv file containing predicted annotations.
        Default is None, in which case the name of the dataset .csv
        is used, with '.annot.csv' appended to it.
    output_dir : str, Path
        path to location where .csv containing predicted annotation
        should be saved. Defaults to current working directory.
    min_segment_dur : float
        minimum duration of segment, in seconds. If specified, then
        any segment with a duration less than min_segment_dur is
        removed from lbl_tb. Default is None, in which case no
        segments are removed.
    majority_vote : bool
        if True, transform segments containing multiple labels
        into segments with a single label by taking a "majority vote",
        i.e. assign all time bins in the segment the most frequently
        occurring label in the segment. This transform can only be
        applied if the labelmap contains an 'unlabeled' label,
        because unlabeled segments makes it possible to identify
        the labeled segments. Default is False.
    save_net_outputs : bool
        if True, save 'raw' outputs of neural networks
        before they are converted to annotations. Default is False.
        Typically the output will be "logits"
        to which a softmax transform might be applied.
        For each item in the dataset--each row in  the `dataset_path` .csv--
        the output will be saved in a separate file in `output_dir`,
        with the extension `{MODEL_NAME}.output.npz`. E.g., if the input is a
        spectrogram with `spect_path` filename `gy6or6_032312_081416.npz`,
        and the network is `TweetyNet`, then the net output file
        will be `gy6or6_032312_081416.tweetynet.output.npz`.
    """
    # ---- pre-conditions ----------------------------------------------------------------------------------------------
    for path, path_name in zip(
        (checkpoint_path, labelmap_path, frames_standardizer_path),
        ("checkpoint_path", "labelmap_path", "frames_standardizer_path"),
    ):
        if path is not None:
            if not validators.is_a_file(path):
                raise FileNotFoundError(
                    f"value for ``{path_name}`` not recognized as a file: {path}"
                )

    model_name = model_config["name"]  # we use this var again below
    if "window_size" not in dataset_config["params"]:
        raise KeyError(
            f"The `dataset_config` for frame classification model '{model_name}' must include a 'params' sub-table "
            f"that sets a value for 'window_size', but received a `dataset_config` that did not:\n{dataset_config}"
        )

    dataset_path = pathlib.Path(dataset_config["path"])
    if not dataset_path.exists() or not dataset_path.is_dir():
        raise NotADirectoryError(
            f"`dataset_path` not found or not recognized as a directory: {dataset_path}"
        )

    # ---- set up directory to save output -----------------------------------------------------------------------------
    # we do this first to make sure we can save things
    if output_dir is None:
        output_dir = pathlib.Path(os.getcwd())
    else:
        output_dir = pathlib.Path(output_dir)

    if not output_dir.is_dir():
        raise NotADirectoryError(
            f"value specified for output_dir is not recognized as a directory: {output_dir}"
        )

    # ---- load what we need to transform data -------------------------------------------------------------------------
    if frames_standardizer_path:
        logger.info(
            f"loading FramesStandardizer from path: {frames_standardizer_path}"
        )
        frames_standardizer = joblib.load(frames_standardizer_path)
    else:
        logger.info("Not loading FramesStandardizer, no path was specified")
        frames_standardizer = None

    logger.info(f"loading labelmap from path: {labelmap_path}")
    with labelmap_path.open("r") as f:
        labelmap = json.load(f)

    # ---------------- load data for prediction ------------------------------------------------------------------------
    if "split" in dataset_config["params"]:
        split = dataset_config["params"]["split"]
        # we do this convoluted thing to avoid 'TypeError: Dataset got multiple values for split`
        del dataset_config["params"]["split"]
    else:
        split = "predict"
    # ---- *not* using a built-in dataset ------------------------------------------------------------------------------
    if dataset_config["name"] is None:
        metadata = datapipes.frame_classification.Metadata.from_dataset_path(
            dataset_path
        )
        dataset_csv_path = dataset_path / metadata.dataset_csv_filename
        metadata = (
            datapipes.frame_classification.metadata.Metadata.from_dataset_path(
                dataset_path
            )
        )
        # we use this below to convert annotations from frames to seconds
        frame_dur = metadata.frame_dur

        logger.info(
            f"loading dataset to predict from csv path: {dataset_csv_path}"
        )

        pred_dataset = InferDatapipe.from_dataset_path(
            dataset_path=dataset_path,
            split=split,
            window_size=dataset_config["params"]["window_size"],
            frames_standardizer=frames_standardizer,
            return_padding_mask=True,
        )
    # ---- *yes* using a built-in dataset ------------------------------------------------------------------------------
    else:
        # we need "target_type" below when converting predictions to annotations,
        # but fail early here if we don't have it
        if "target_type" not in dataset_config["params"]:
            from ..datasets.biosoundsegbench import VALID_TARGET_TYPES

            raise ValueError(
                "The dataset table in the configuration file requires a 'target_type' "
                "when running predictions on built-in datasets. "
                "Please add a key to the table whose value is a valid target type: "
                f"{VALID_TARGET_TYPES}"
            )
        dataset_config["params"]["return_padding_mask"] = True
        # next line, required to be true regardless of split so we set it here
        dataset_config["params"]["return_frames_path"] = True
        pred_dataset = datasets.get(
            dataset_config,
            split=split,
            frames_standardizer=frames_standardizer,
        )
        # we use this below to convert annotations from frames to seconds
        frame_dur = pred_dataset.frame_dur

    pred_loader = torch.utils.data.DataLoader(
        dataset=pred_dataset,
        shuffle=False,
        # batch size 1 because each spectrogram reshaped into a batch of windows
        batch_size=1,
        num_workers=num_workers,
    )

    logger.info(
        f"Duration of a frame in dataset, in seconds: {frame_dur}",
    )

    # ---------------- do the actual predicting + converting to annotations --------------------------------------------
    input_shape = pred_dataset.shape
    # if dataset returns spectrogram reshaped into windows,
    # throw out the window dimension; just want to tell network (channels, height, width) shape
    if len(input_shape) == 4:
        input_shape = input_shape[1:]
    logger.info(
        f"Shape of input to networks used for predictions: {input_shape}"
    )

    logger.info(f"instantiating model from config:/n{model_name}")

    model = models.get(
        model_name,
        model_config,
        num_classes=len(labelmap),
        input_shape=input_shape,
        labelmap=labelmap,
    )

    # ---------------- do the actual predicting --------------------------------------------------------------------
    logger.info(
        f"loading checkpoint for {model_name} from path: {checkpoint_path}"
    )
    model.load_state_dict_from_path(checkpoint_path)

    trainer_logger = lightning.pytorch.loggers.TensorBoardLogger(
        save_dir=output_dir
    )
    trainer = lightning.pytorch.Trainer(
        accelerator=trainer_config["accelerator"],
        devices=trainer_config["devices"],
        logger=trainer_logger,
    )

    logger.info(f"running predict method of {model_name}")
    results = trainer.predict(model, pred_loader)
    # TODO: figure out how to overload `on_predict_epoch_end` to return dict
    pred_dict = {
        frames_path: y_pred
        for result in results
        for frames_path, y_pred in result.items()
    }

    # ---------------- set up to convert predictions to annotation files -----------------------------------------------
    if dataset_config["name"] is None:
        # we assume this default for now -- prep'd datasets are always multi-class frame label
        target_type = "multi_frame_labels"
    else:
        # we made sure we have this above when determining the kind of dataset
        target_type = dataset_config["params"]["target_type"]
        if isinstance(target_type, str):
            pass
        elif isinstance(target_type, (list, tuple)):
            target_type = tuple(sorted(target_type))

    if annot_csv_filename is None:
        annot_csv_filename = (
            pathlib.Path(dataset_path).stem + constants.ANNOT_CSV_SUFFIX
        )
    annot_csv_path = pathlib.Path(output_dir).joinpath(annot_csv_filename)
    logger.info(f"will save annotations in .csv file: {annot_csv_path}")

    # ----------------  converting to annotations ------------------------------------------------------------------
    progress_bar = tqdm(pred_loader)

    if dataset_config["name"] is None:
        # we're using a user-prepped dataset, not a built-in dataset
        # so assume we have metadata from above
        input_type = (
            metadata.input_type
        )  # we use this to get frame_times inside loop
        if input_type == "audio":
            audio_format = metadata.audio_format
        elif input_type == "spect":
            spect_format = metadata.spect_format
    else:
        input_type = "spect"  # assume this for now
        spect_format = common.constants.DEFAULT_SPECT_FORMAT

    annots = []
    logger.info("converting predictions to annotations")
    for ind, batch in enumerate(progress_bar):
        padding_mask, frames_path = batch["padding_mask"], batch["frames_path"]
        padding_mask = np.squeeze(padding_mask)
        if isinstance(frames_path, list) and len(frames_path) == 1:
            frames_path = frames_path[0]
        # we do all this basically to have clear naming below
        if (
            target_type == "multi_frame_labels"
            or target_type == "binary_frame_labels"
        ):
            class_logits = pred_dict[frames_path]
            boundary_logits = None
        elif target_type == "boundary_frame_labels":
            boundary_logits = pred_dict[frames_path]
            class_logits = None
        elif target_type == ("boundary_frame_labels", "multi_frame_labels"):
            class_logits, boundary_logits = pred_dict[frames_path]

        if save_net_outputs:
            # not sure if there's a better way to get outputs into right shape;
            # can't just call y_pred.reshape() because that basically flattens the whole array first
            # meaning we end up with elements in the wrong order
            # so instead we convert to sequence then stack horizontally, on column axis
            net_output = torch.hstack(class_logits.unbind())
            net_output = net_output[:, padding_mask]
            net_output = net_output.cpu().numpy()
            net_output_path = output_dir.joinpath(
                pathlib.Path(frames_path).stem
                + f"{model_name}{constants.NET_OUTPUT_SUFFIX}"
            )
            np.savez(net_output_path, net_output)

        if class_logits is not None:
            class_preds = torch.argmax(
                class_logits, dim=1
            )  # assumes class dimension is 1
            class_preds = (
                torch.flatten(class_preds).cpu().numpy()[padding_mask]
            )
        if boundary_logits is not None:
            boundary_preds = torch.argmax(
                boundary_logits, dim=1
            )  # assumes class dimension is 1
            boundary_preds = (
                torch.flatten(boundary_preds).cpu().numpy()[padding_mask]
            )

        if input_type == "audio":
            frames, samplefreq = constants.AUDIO_FORMAT_FUNC_MAP[audio_format](
                frames_path
            )
            frame_times = np.arange(frames.shape[-1]) / samplefreq
        elif input_type == "spect":
            spect_dict = files.spect.load(
                frames_path, spect_format=spect_format
            )
            frame_times = spect_dict[timebins_key]

        # audio_fname is used for audio_path attribute of crowsetta.Annotation below
        audio_fname = files.spect.find_audio_fname(frames_path)
        if (
            target_type == "multi_frame_labels"
            or target_type == "binary_frame_labels"
        ):
            if majority_vote or min_segment_dur:
                if background_label in labelmap:
                    background_label = labelmap[background_label]
                elif (
                    "unlabeled" in labelmap
                ):  # some backward compatibility here
                    background_label = labelmap["unlabeled"]
                else:
                    background_label = 0  # set a default value anyway just to not throw an error
                class_preds = transforms.frame_labels.postprocess(
                    class_preds,
                    timebin_dur=frame_dur,
                    min_segment_dur=min_segment_dur,
                    majority_vote=majority_vote,
                    background_label=background_label,
                )
            labels, onsets_s, offsets_s = transforms.frame_labels.to_segments(
                class_preds,
                labelmap=labelmap,
                frame_times=frame_times,
            )
            if labels is None and onsets_s is None and offsets_s is None:
                # handle the case when all time bins are predicted to be unlabeled
                # see https://github.com/NickleDave/vak/issues/383
                continue
            seq = crowsetta.Sequence.from_keyword(
                labels=labels, onsets_s=onsets_s, offsets_s=offsets_s
            )

            annot = crowsetta.Annotation(
                seq=seq,
                notated_path=audio_fname,
                annot_path=annot_csv_path.name,
            )
            annots.append(annot)

        elif target_type == "boundary_frame_labels":
            boundary_inds = (
                transforms.frame_labels.boundary_inds_from_boundary_labels(
                    boundary_preds,
                    force_boundary_first_ind=True,
                )
            )
            boundary_times = frame_times[boundary_inds]  # fancy indexing
            df = pd.DataFrame.from_records({"boundary_time": boundary_times})
            annots.append(AnnotationDataFrame(df=df, audio_path=audio_fname))
        elif target_type == ("boundary_frame_labels", "multi_frame_labels"):
            if majority_vote is False:
                logger.warn(
                    "`majority_vote` was set to False but `vak.predict.predict_with_frame_classification_model` "
                    "determined that this model predicts both multi-class labels and boundary labels, "
                    "so `majority_vote` will be set to True (to assign a single label to each segment determined by "
                    "a boundary)"
                )
            if background_label in labelmap:
                background_label = labelmap[background_label]
            elif "unlabeled" in labelmap:  # some backward compatibility here
                background_label = labelmap["unlabeled"]
            else:
                background_label = (
                    0  # set a default value anyway just to not throw an error
                )
            # Notice here we *always* call post-process, with majority_vote=True
            # because we are using boundary labels
            class_preds = transforms.frame_labels.postprocess(
                frame_labels=class_preds,
                timebin_dur=frame_dur,
                min_segment_dur=min_segment_dur,
                majority_vote=True,
                background_label=background_label,
                boundary_labels=boundary_preds,
            )
            labels, onsets_s, offsets_s = transforms.frame_labels.to_segments(
                class_preds,
                labelmap=labelmap,
                frame_times=frame_times,
            )
            if labels is None and onsets_s is None and offsets_s is None:
                # handle the case when all time bins are predicted to be unlabeled
                # see https://github.com/NickleDave/vak/issues/383
                continue
            seq = crowsetta.Sequence.from_keyword(
                labels=labels, onsets_s=onsets_s, offsets_s=offsets_s
            )

            annot = crowsetta.Annotation(
                seq=seq,
                notated_path=audio_fname,
                annot_path=annot_csv_path.name,
            )
            annots.append(annot)
    if len(annots) < 1:
        # catch edge case where nothing was predicted
        # FIXME: this should have columns that match GenericSeq
        pd.DataFrame.from_records([]).to_csv(annot_csv_path)
    elif all([isinstance(annot, crowsetta.Annotation) for annot in annots]):
        generic_seq = crowsetta.formats.seq.GenericSeq(annots=annots)
        generic_seq.to_file(annot_path=annot_csv_path)
    elif all([isinstance(annot, AnnotationDataFrame) for annot in annots]):
        df_out = []
        for sample_num, annot_df in enumerate(annots):
            df = annot_df.df
            df["audio_path"] = str(annot_df.audio_path)
            df["sample_num"] = sample_num
            df_out.append(df)
        df_out = pd.concat(df_out)
        df_out.to_csv(annot_csv_path, index=False)



"""Tests for vak.train.train function."""
from unittest import mock

import pytest

import vak.config
import vak.common.constants
import vak.common.paths
import vak.train


@pytest.mark.parametrize(
    "audio_format, spect_format, annot_format, model_name, train_function_to_mock",
    [
        ("cbin", None, "notmat", "TweetyNet",
         'vak.train.train_.train_frame_classification_model'),
        (None, "mat", "yarden", "TweetyNet",
         'vak.train.train_.train_frame_classification_model'),
        ("cbin", None, "notmat", "ConvEncoderUMAP",
         'vak.train.train_.train_parametric_umap_model'),
    ],
)
def test_train(
    audio_format, spect_format, annot_format, model_name, train_function_to_mock,
        specific_config_toml_path, tmp_path
):
    """Test that :func:`vak.train.train` dispatches to the correct model-specific
    training functions"""
    root_results_dir = tmp_path.joinpath("test_train_root_results_dir")
    root_results_dir.mkdir()

    keys_to_change = [
        {
            "table": "train",
            "key": "root_results_dir",
            "value": str(root_results_dir),
        },
        {"table": "train", "key": "trainer", "value": {"accelerator": "cpu", "devices": 1}},
    ]

    toml_path = specific_config_toml_path(
        config_type="train",
        model=model_name,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)

    results_path = tmp_path / 'results_path'
    results_path.mkdir()

    with mock.patch(train_function_to_mock, autospec=True) as mock_train_function:
        vak.train.train(
            model_config=cfg.train.model.asdict(),
            dataset_config=cfg.train.dataset.asdict(),
            trainer_config=cfg.train.trainer.asdict(),
            batch_size=cfg.train.batch_size,
            num_epochs=cfg.train.num_epochs,
            num_workers=cfg.train.num_workers,
            checkpoint_path=cfg.train.checkpoint_path,
            frames_standardizer_path=cfg.train.frames_standardizer_path,
            results_path=results_path,
            standardize_frames=cfg.train.standardize_frames,
            shuffle=cfg.train.shuffle,
            val_step=cfg.train.val_step,
            ckpt_step=cfg.train.ckpt_step,
            patience=cfg.train.patience,
        )
        assert mock_train_function.called



"""High-level function that generates new inferences from trained models."""

from __future__ import annotations

import logging
import os
import pathlib

from .. import models
from ..common import validators
from ..common.accelerator import get_default as get_default_device
from .frame_classification import predict_with_frame_classification_model

logger = logging.getLogger(__name__)


def predict(
    model_config: dict,
    dataset_config: dict,
    trainer_config: dict,
    checkpoint_path: str | pathlib.Path,
    labelmap_path: str | pathlib.Path,
    num_workers: int = 2,
    timebins_key: str = "t",
    frames_standardizer_path: str | pathlib.Path | None = None,
    device: str | None = None,
    annot_csv_filename: str | None = None,
    output_dir: str | pathlib.Path | None = None,
    min_segment_dur: float | None = None,
    majority_vote: bool = False,
    save_net_outputs: bool = False,
):
    """Make predictions on a dataset with a trained model.

    Parameters
    ----------
    model_config : dict
        Model configuration in a ``dict``,
        as loaded from a .toml file,
        and used by the model method ``from_config``.
    dataset_config: dict
        Dataset configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.DatasetConfig.asdict`.
    trainer_config: dict
        Configuration for :class:`lightning.pytorch.Trainer`.
        Can be obtained by calling :meth:`vak.config.TrainerConfig.asdict`.
    checkpoint_path : str
        path to directory with checkpoint files saved by Torch, to reload model
    labelmap_path : str
        path to 'labelmap.json' file.
    window_size : int
        size of windows taken from spectrograms, in number of time bins,
        shown to neural networks
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader. Default is 2.
    timebins_key : str
        key for accessing vector of time bins in files. Default is 't'.
    frames_standardizer_path : str
        path to a saved :class:`vak.transforms.FramesStandardizer` object used to standardize (normalize) frames.
        If spectrograms were normalized and this is not provided, will give
        incorrect results.
    annot_csv_filename : str
        name of .csv file containing predicted annotations.
        Default is None, in which case the name of the dataset .csv
        is used, with '.annot.csv' appended to it.
    output_dir : str, Path
        path to location where .csv containing predicted annotation
        should be saved. Defaults to current working directory.
    min_segment_dur : float
        minimum duration of segment, in seconds. If specified, then
        any segment with a duration less than min_segment_dur is
        removed from lbl_tb. Default is None, in which case no
        segments are removed.
    majority_vote : bool
        if True, transform segments containing multiple labels
        into segments with a single label by taking a "majority vote",
        i.e. assign all time bins in the segment the most frequently
        occurring label in the segment. This transform can only be
        applied if the labelmap contains an 'unlabeled' label,
        because unlabeled segments makes it possible to identify
        the labeled segments. Default is False.
    save_net_outputs : bool
        If True, save 'raw' outputs of neural networks
        before they are converted to annotations. Default is False.
        Typically the output will be "logits"
        to which a softmax transform might be applied.
        For each item in the dataset--each row in  the `dataset_path` .csv--
        the output will be saved in a separate file in `output_dir`,
        with the extension `{MODEL_NAME}.output.npz`. E.g., if the input is a
        spectrogram with `spect_path` filename `gy6or6_032312_081416.npz`,
        and the network is `TweetyNet`, then the net output file
        will be `gy6or6_032312_081416.tweetynet.output.npz`.
    """
    for path, path_name in zip(
        (checkpoint_path, labelmap_path, frames_standardizer_path),
        ("checkpoint_path", "labelmap_path", "frames_standardizer_path"),
    ):
        if path is not None:
            if not validators.is_a_file(path):
                raise FileNotFoundError(
                    f"value for ``{path_name}`` not recognized as a file: {path}"
                )

    dataset_path = pathlib.Path(dataset_config["path"])
    if not dataset_path.exists() or not dataset_path.is_dir():
        raise NotADirectoryError(
            f"`dataset_path` not found or not recognized as a directory: {dataset_path}"
        )

    if output_dir is None:
        output_dir = pathlib.Path(os.getcwd())
    else:
        output_dir = pathlib.Path(output_dir)

    if not output_dir.is_dir():
        raise NotADirectoryError(
            f"value specified for output_dir is not recognized as a directory: {output_dir}"
        )

    if device is None:
        device = get_default_device()

    model_name = model_config["name"]
    try:
        model_family = models.registry.MODEL_FAMILY_FROM_NAME[model_name]
    except KeyError as e:
        raise ValueError(
            f"No model family found for the model name specified: {model_name}"
        ) from e
    if model_family == "FrameClassificationModel":
        predict_with_frame_classification_model(
            model_config=model_config,
            dataset_config=dataset_config,
            trainer_config=trainer_config,
            checkpoint_path=checkpoint_path,
            labelmap_path=labelmap_path,
            num_workers=num_workers,
            timebins_key=timebins_key,
            frames_standardizer_path=frames_standardizer_path,
            annot_csv_filename=annot_csv_filename,
            output_dir=output_dir,
            min_segment_dur=min_segment_dur,
            majority_vote=majority_vote,
            save_net_outputs=save_net_outputs,
        )
    else:
        raise ValueError(f"Model family not recognized: {model_family}")



from __future__ import annotations

import pathlib

import numpy as np
import pandas as pd

from . import annotation, constants


def to_map(
    labelset: set,
    map_background: bool = True,
    background_label: str = constants.DEFAULT_BACKGROUND_LABEL,
) -> dict:
    """Convert set of labels to `dict`
    mapping those labels to a series of consecutive integers
    from 0 to n inclusive,
    where n is the number of labels in the set.

    This 'labelmap' is used when mapping labels
    from annotations of a vocalization into
    a label for every time bin in a spectrogram of that vocalization.

    If ``map_background`` is True, then a label
    will be added to labelset representing a background class
    (any segment that is not labeled).
    The default for this label is
    :const:`vak.common.constants.DEFAULT_BACKGROUND_LABEL`.
    This string label will map to class index 0,
    so the total number of classes is n + 1.

    Parameters
    ----------
    labelset : set
        Set of labels used to annotate a dataset.
    map_background : bool
        If True, include key specified by
        ``background_label`` in mapping.
        Any time bins in a spectrogram
        that do not have a label associated with them,
        e.g. a silent gap between vocalizations,
        will be assigned the integer
        that the background key maps to.
    background_label: str, optional
        The string label applied to segments belonging to the
        background class.
        Default is
        :const:`vak.common.constants.DEFAULT_BACKGROUND_LABEL`.

    Returns
    -------
    labelmap : dict
        Maps labels to integers.
    """
    if not isinstance(labelset, set):
        raise TypeError(
            f"type of labelset must be set, got type {type(labelset)}"
        )

    labellist = []
    if map_background is True:
        # NOTE we append background label *first*
        labellist.append(background_label)
    # **then** extend with the rest of the labels
    labellist.extend(sorted(list(labelset)))
    # so that background_label maps to class index 0 by default in next line
    labelmap = dict(zip(labellist, range(len(labellist))))
    return labelmap


def to_set(labels_list: list[np.ndarray | list]) -> set:
    """Given a list of labels from annotations,
    return the set of (unique) labels.

    Parameters
    ----------
    labels_list : list
         Of labels from annotations,
         either a list of numpy.ndarrays
         or a list of lists.

    Returns
    -------
    labelset : set
        Unique set of labels found in ``labels_list``.

    Examples
    --------
    >>> labels_list = [voc.annot.labels for voc in vds.voc_list]
    >>> labelset = to_set(labels_list)
    >>> print(labelset)
    {'a', 'b', 'c', 'd', 'e'}
    """
    all_labels = [lbl for labels in labels_list for lbl in labels]
    labelset = set(all_labels)
    return labelset


def from_df(
    dataset_df: pd.DataFrame, dataset_path: str | pathlib.Path
) -> list[np.ndarray]:
    """Returns labels for each vocalization in a dataset.

    Takes Pandas DataFrame representing the dataset, loads
    annotation for each row in the DataFrame, and then returns
    labels from each annotation.

    Parameters
    ----------
    dataset_df : pandas.DataFrame
        created by vak.io.dataframe.from_files

    Returns
    -------
    labels : list
        of array-like, labels for each vocalization in the dataset.
    """
    dataset_path = pathlib.Path(dataset_path)
    if not dataset_path.exists() or not dataset_path.is_dir():
        raise NotADirectoryError(
            f"`dataset_path` not found or not recognized as a directory: {dataset_path}"
        )

    annots = annotation.from_df(dataset_df, dataset_path)
    return [annot.seq.labels for annot in annots]


ALPHANUMERIC = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"
DUMMY_SINGLE_CHAR_LABELS = [
    chr(x)
    # some large range of characters not typically used as labels
    for x in range(162, 2000)
]
# start with alphanumeric since more human readable;
# mapping can be arbitrary as long as it's consistent
DUMMY_SINGLE_CHAR_LABELS = (*ALPHANUMERIC, *DUMMY_SINGLE_CHAR_LABELS)


# added to fix https://github.com/NickleDave/vak/issues/373
def multi_char_labels_to_single_char(
    labelmap: dict, skip: tuple[str] = (constants.DEFAULT_BACKGROUND_LABEL,)
) -> dict:
    """Return a copy of a ``labelmap`` where any
    labels that are strings with multiple characters
    are converted to single characters.

    This makes it possible to correctly compute metrics
    like Levenshtein edit distance.

    Labels that are strings with multiple characters
    are replaced by a single-label character from
    the constant ``vak.labels.DUMMY_SINGLE_CHAR_LABELS``.
    The replacement is grabbed with the index of the
    multi-character label from the sorted ``dict``.

    Parameters
    ----------
    labelmap : dict
        That maps human-readable string labels
        to integers. As returned by
        ``vak.labels.to_map``.
    skip : tuple
        A tuple of labels to leave as multiple characters.
        Default is a tuple containing just
        :const:`vak.common.constants.DEFAULT_BACKGROUND_LABEL`.

    Returns
    -------
    labelmap : dict
        Where any keys with multiple characters
        in string are converted to dummy single characters.
    """
    current_str_labels = sorted(
        # sort to be extra sure we get same order every time
        # (even though OrderedDict is now default in Python).
        # Same order forces mapping to single characters to be deterministic across function calls.
        labelmap.keys()
    )
    if all([len(lbl) == 1 for lbl in current_str_labels]):
        # no need to do re-mapping
        return labelmap

    # We only use single character labels that are not already in labelmap,
    # to avoid over-writing a single-character label from the original labelmap
    # with the same single-character from DUMMY_SINGLE_CHAR_LABELS,
    # which would map it to a new integer and cause us to lose the original integer
    # from the mapping
    single_char_labels_not_in_labelmap = [
        lbl for lbl in DUMMY_SINGLE_CHAR_LABELS if lbl not in labelmap
    ]
    n_needed_to_remap = len(
        [lbl for lbl in current_str_labels if len(lbl) > 1]
    )
    if n_needed_to_remap > len(single_char_labels_not_in_labelmap):
        raise ValueError(
            f"Need to remap {n_needed_to_remap} multiple-character labels"
            f"but there are only {len(single_char_labels_not_in_labelmap)} available."
        )

    new_labelmap = {}
    for dummy_label_ind, label_str in enumerate(current_str_labels):
        label_int = labelmap[label_str]
        if (
            len(label_str) > 1 and label_str not in skip
        ):  # default for `skip` is ('unlabeled',)
            # replace with dummy label
            new_label_str = single_char_labels_not_in_labelmap[dummy_label_ind]
            new_labelmap[new_label_str] = label_int
        else:
            new_labelmap[label_str] = label_int
    return new_labelmap



import numpy as np
import pytest

import vak.datapipes.frame_classification.helper

from ... import fixtures


@pytest.mark.parametrize(
    'subset',
    [
        'train-dur-4.0-replicate-1',
        'train-dur-4.0-replicate-2'
    ]
)
def test_sample_ids_array_filename_for_subset(subset):
    out = vak.datapipes.frame_classification.helper.sample_ids_array_filename_for_subset(subset)
    assert isinstance(out, str)
    assert out == vak.datapipes.frame_classification.constants.SAMPLE_IDS_ARRAY_FILENAME.replace(
                '.npy', f'-{subset}.npy'
            )


@pytest.mark.parametrize(
    'subset',
    [
        'train-dur-4.0-replicate-1',
        'train-dur-4.0-replicate-2'
    ]
)
def test_inds_in_sample_array_filename_for_subset(subset):
    out = vak.datapipes.frame_classification.helper.inds_in_sample_array_filename_for_subset(subset)
    assert isinstance(out, str)
    assert out == vak.datapipes.frame_classification.constants.INDS_IN_SAMPLE_ARRAY_FILENAME.replace(
                '.npy', f'-{subset}.npy'
            )


@pytest.fixture(params=fixtures.spect.SPECT_LIST_NPZ)
def frames_path(request):
    return request.param


def test_load_frames(frames_path):
    out = vak.datapipes.frame_classification.helper.load_frames(frames_path, input_type="spect")
    assert isinstance(out, np.ndarray)



"""fixtures relating to array files containing spectrograms"""
import pytest

import vak.common.files.spect

from .annot import (
    ANNOT_LIST_NOTMAT,
    ANNOT_LIST_YARDEN,
    LABELSET_NOTMAT,
    LABELSET_YARDEN,
)
from .test_data import GENERATED_TEST_DATA_ROOT, SOURCE_TEST_DATA_ROOT


SPECT_DIR_MAT = SOURCE_TEST_DATA_ROOT.joinpath(
    "spect_mat_annot_yarden", "llb3", "spect"
)


@pytest.fixture
def spect_dir_mat():
    return SPECT_DIR_MAT


# this dir is created by ./tests/scripts/generate_test_data
SPECT_DIR_NPZ = GENERATED_TEST_DATA_ROOT / "spect-output-dir/audio_cbin_annot_notmat/gy6or6/032312"


@pytest.fixture
def spect_dir_npz():
    return SPECT_DIR_NPZ


@pytest.fixture
def specific_spect_dir(spect_dir_mat, spect_dir_npz):
    def _specific_spect_dir(spect_format):
        if spect_format == "mat":
            return spect_dir_mat
        elif spect_format == "npz":
            return spect_dir_npz
        else:
            raise ValueError(f"invalid spect_format: {spect_format}")

    return _specific_spect_dir


SPECT_LIST_MAT = sorted(SPECT_DIR_MAT.glob("*.mat"))


@pytest.fixture
def spect_list_mat():
    return SPECT_LIST_MAT


SPECT_LIST_NPZ = sorted(SPECT_DIR_NPZ.glob("*.spect.npz"))


@pytest.fixture
def spect_list_npz():
    return SPECT_LIST_NPZ


LABELSET_YARDEN_SET = set(LABELSET_YARDEN)
SPECT_LIST_MAT_ALL_LABELS_IN_LABELSET = []
SPECT_LIST_MAT_LABELS_NOT_IN_LABELSET = []
for spect_path in SPECT_LIST_MAT:
    audio_fname = vak.common.files.spect.find_audio_fname(spect_path)
    annot = [
        annot for annot in ANNOT_LIST_YARDEN if annot.notated_path.name == audio_fname
    ]
    assert len(annot) == 1
    annot = annot[0]
    if set(annot.seq.labels).issubset(LABELSET_YARDEN_SET):
        SPECT_LIST_MAT_ALL_LABELS_IN_LABELSET.append(spect_path)
    else:
        SPECT_LIST_MAT_LABELS_NOT_IN_LABELSET.append(spect_path)


@pytest.fixture
def spect_list_mat_all_labels_in_labelset():
    """list of .mat spectrogram files where all labels in associated annotation **are** in labelset"""
    return SPECT_LIST_MAT_ALL_LABELS_IN_LABELSET


@pytest.fixture
def spect_list_mat_labels_not_in_labelset():
    """list of .mat spectrogram files where some labels in associated annotation are **not** in labelset"""
    return SPECT_LIST_MAT_LABELS_NOT_IN_LABELSET


LABELSET_NOTMAT_SET = set(LABELSET_NOTMAT)
SPECT_LIST_NPZ_ALL_LABELS_IN_LABELSET = []
SPECT_LIST_NPZ_LABELS_NOT_IN_LABELSET = []
for spect_path in SPECT_LIST_NPZ:
    audio_fname = vak.common.files.spect.find_audio_fname(spect_path)
    annot = [
        annot for annot in ANNOT_LIST_NOTMAT if annot.notated_path.name == audio_fname
    ]
    assert len(annot) == 1
    annot = annot[0]
    if set(annot.seq.labels).issubset(LABELSET_NOTMAT_SET):
        SPECT_LIST_NPZ_ALL_LABELS_IN_LABELSET.append(spect_path)
    else:
        SPECT_LIST_NPZ_LABELS_NOT_IN_LABELSET.append(spect_path)


@pytest.fixture
def spect_list_npz_all_labels_in_labelset():
    """list of .npz spectrogram files where all labels in associated annotation **are** in labelset"""
    return SPECT_LIST_NPZ_ALL_LABELS_IN_LABELSET


@pytest.fixture
def spect_list_npz_labels_not_in_labelset():
    """list of .npz spectrogram files where some labels in associated annotation are  **not** in labelset"""
    return SPECT_LIST_NPZ_LABELS_NOT_IN_LABELSET


@pytest.fixture
def specific_spect_list(
    spect_list_mat,
    spect_list_mat_all_labels_in_labelset,
    spect_list_mat_labels_not_in_labelset,
    spect_list_npz,
    spect_list_npz_all_labels_in_labelset,
    spect_list_npz_labels_not_in_labelset,
):
    def _specific_spect_list(spect_format, qualifier=None):
        MAP = {
            "mat": {
                None: spect_list_mat,
                "all_labels_in_labelset": spect_list_mat_all_labels_in_labelset,
                "labels_not_in_labelset": spect_list_mat_labels_not_in_labelset,
            },
            "npz": {
                None: spect_list_npz,
                "all_labels_in_labelset": spect_list_npz_all_labels_in_labelset,
                "labels_not_in_labelset": spect_list_npz_labels_not_in_labelset,
            },
        }
        return MAP[spect_format][qualifier]

    return _specific_spect_list



"""Decorator that makes a :class:`vak.models.ModelFactory`,
given a definition of the model,
and a :class:`lightning.LightningModule` that represents a
family of models that the new model belongs to.

The function returns a new instance of :class:`vak.models.ModelFactory`,
that can create new instances of the model with its
:meth:`~:class:`vak.models.ModelFactory.from_config` and
:meth:`~:class:`vak.models.ModelFactory.from_instances` methods.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Type

import lightning

from .registry import register_model

if TYPE_CHECKING:
    from .factory import ModelFactory


class ModelDefinitionValidationError(Exception):
    """Exception raised when validating a model
    definition fails.

    Used by :func:`vak.models.decorator.model` decorator.
    """

    pass


def model(family: lightning.pytorch.LightningModule):
    """Decorator that makes a :class:`vak.models.ModelFactory`,
    given a definition of the model,
    and a :class:`lightning.LightningModule` that represents a
    family of models that the new model belongs to.

    The function returns a new instance of :class:`vak.models.ModelFactory`,
    that can create new instances of the model with its
    :meth:`~:class:`vak.models.ModelFactory.from_config` and
    :meth:`~:class:`vak.models.ModelFactory.from_instances` methods.

    Parameters
    ----------
    definition : type
        The definition of the new model that will be made.
        A class with all the class variables required
        by :func:`vak.models.definition.validate`.
        See docstring of that function for specification.
        See also :class:`vak.models.definition.ModelDefinition`,
        but note that it is not necessary to subclass
        :class:`~vak.models.definition.ModelDefinition` to
        define a model.
    family : lightning.LightningModule
        The class representing the family of models
        that the new model will belong to.
        E.g., :class:`vak.models.FrameClassificationModel`.
        Should be a subclass of :class:`lightning.LightningModule`
        that was registered with the
        :func:`vak.models.registry.model_family` decorator.

    Returns
    -------
    model_factory : vak.models.ModelFactory
        An instance of :class:`~vak.models.ModelFactory`,
        with attribute ``definition`` and ``family``,
        that will be used when making
        new instances of the model by calling the
        :meth:`~vak.models.ModelFactory.from_config` method
        or the :meth:`~:class:`vak.models.ModelFactory.from_instances` method.
    """

    def _model(definition: Type) -> ModelFactory:
        from .factory import ModelFactory  # avoid circular import

        model_factory = ModelFactory(definition, family)
        model_factory.__name__ = definition.__name__
        model_factory.__doc__ = definition.__doc__
        model_factory.__module__ = definition.__module__
        register_model(model_factory)
        return model_factory

    return _model



"""High-level function that generates results for a learning curve for all models."""

from __future__ import annotations

import logging
import pathlib

from .. import models
from ..common.converters import expanded_user_path
from .frame_classification import learning_curve_for_frame_classification_model

logger = logging.getLogger(__name__)


def learning_curve(
    model_config: dict,
    dataset_config: dict,
    trainer_config: dict,
    batch_size: int,
    num_epochs: int,
    num_workers: int,
    results_path: str | pathlib.Path = None,
    post_tfm_kwargs: dict | None = None,
    standardize_frames: bool = True,
    shuffle: bool = True,
    val_step: int | None = None,
    ckpt_step: int | None = None,
    patience: int | None = None,
) -> None:
    """Generate results for a learning curve,
    where model performance is measured as a
    function of dataset duration in seconds.

    Trains a class of model with a range of dataset durations,
    and then evaluates each trained model
    with a test set that is held constant
    (unlike the training sets).
    Results are saved in a new directory within ``root_results_dir``.

    Parameters
    ----------
    model_config : dict
        Model configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.ModelConfig.asdict`.
    dataset_config: dict
        Dataset configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.DatasetConfig.asdict`.
    trainer_config: dict
        Configuration for :class:`lightning.pytorch.Trainer`.
        Can be obtained by calling :meth:`vak.config.TrainerConfig.asdict`.
    batch_size : int
        number of samples per batch presented to models during training.
    num_epochs : int
        number of training epochs. One epoch = one iteration through the entire
        training set.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader.
    results_path : str, pathlib.Path
        Directory where results will be saved.
    previous_run_path : str, pathlib.Path
        path to directory containing dataset .csv files
        that represent subsets of training set, created by
        a previous run of ``vak.core.learncurve.learning_curve``.
        Typically directory will have a name like ``results_{timestamp}``
        and the actual .csv splits will be in sub-directories with names
        corresponding to the training set duration
    post_tfm_kwargs : dict
        Keyword arguments to post-processing transform.
        If None, then no additional clean-up is applied
        when transforming labeled timebins to segments,
        the default behavior. The transform used is
        ``vak.transforms.frame_labels.ToSegmentsWithPostProcessing`.
        Valid keyword argument names are 'majority_vote'
        and 'min_segment_dur', and should be appropriate
        values for those arguments: Boolean for ``majority_vote``,
        a float value for ``min_segment_dur``.
        See the docstring of the transform for more details on
        these arguments and how they work.
    shuffle: bool
        if True, shuffle training data before each epoch. Default is True.
    standardize_frames : bool
        if True, use :class:`vak.transforms.FramesStandardizer` to standardize the frames.
        Normalization is done by subtracting off the mean for each row
        of the training set and then dividing by the std for that frequency bin.
        This same normalization is then applied to validation + test data.
    val_step : int
        Step on which to estimate accuracy using validation set.
        If val_step is n, then validation is carried out every time
        the global step / n is a whole number, i.e., when val_step modulo the global step is 0.
        Default is None, in which case no validation is done.
    ckpt_step : int
        Step on which to save to checkpoint file.
        If ckpt_step is n, then a checkpoint is saved every time
        the global step / n is a whole number, i.e., when ckpt_step modulo the global step is 0.
        Default is None, in which case checkpoint is only saved at the last epoch.
    patience : int
        number of validation steps to wait without performance on the
        validation set improving before stopping the training.
        Default is None, in which case training only stops after the specified number of epochs.
    """
    # ---------------- pre-conditions ----------------------------------------------------------------------------------
    dataset_path = expanded_user_path(dataset_config["path"])
    if not dataset_path.exists() or not dataset_path.is_dir():
        raise NotADirectoryError(
            f"`dataset_path` not found or not recognized as a directory: {dataset_path}"
        )

    model_name = model_config["name"]
    try:
        model_family = models.registry.MODEL_FAMILY_FROM_NAME[model_name]
    except KeyError as e:
        raise ValueError(
            f"No model family found for the model name specified: {model_name}"
        ) from e
    if model_family == "FrameClassificationModel":
        learning_curve_for_frame_classification_model(
            model_config=model_config,
            dataset_config=dataset_config,
            trainer_config=trainer_config,
            batch_size=batch_size,
            num_epochs=num_epochs,
            num_workers=num_workers,
            results_path=results_path,
            post_tfm_kwargs=post_tfm_kwargs,
            standardize_frames=standardize_frames,
            shuffle=shuffle,
            val_step=val_step,
            ckpt_step=ckpt_step,
            patience=patience,
        )
    else:
        raise ValueError(f"Model family not recognized: {model_family}")



"""Modules that act as activation functions."""

import torch


class NormReLU(torch.nn.Module):
    """This module implements normalized ReLU activation,
    as used with the Encoder-Decoder Temporal Convolutional Model in [1]_.

    The output of the activation is normalized by the channel-wise maximum.

    .. [1] Lea, C., Flynn, M. D., Vidal, R., Reiter, A., & Hager, G. D. (2017).
       Temporal convolutional networks for action segmentation and detection.
       In proceedings of the IEEE Conference on Computer Vision and Pattern Recognition (pp. 156-165).
    """

    def __init__(self):
        super().__init__()
        self.relu = torch.nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.relu(x)
        max_values = torch.max(torch.abs(x), dim=-1, keepdims=True)[0] + 1e-5
        x = x / max_values
        return x
